# This module provides routines for installing the Crystal application
# into the Linux desktop environment (e.g., creating .desktop files,
# MIME types, and icons). This module is invoked when the
# --install-to-desktop command-line argument is provided to Crystal.
# 
# This module is only intended to be used on Linux systems and should not be
# imported or executed on non-Linux platforms.

from collections.abc import Iterator
from contextlib import contextmanager
from crystal.util.ellipsis import Ellipsis, EllipsisType
from functools import cache
import os
import os.path
import shutil
import subprocess
import sys
from typing import BinaryIO, Dict, Literal, Optional, Tuple, Union
import xml.etree.ElementTree as ET


def install_to_linux_desktop_environment() -> None:
    """
    Install the Crystal application to the desktop environment, in Linux.
    """
    from crystal import resources
    from crystal.util import gio
    from crystal.util.xos import is_linux
    import glob
    
    if not is_linux():
        raise ValueError()
    
    # Ensure not running as root (unless is actually the root user)
    running_as_root_user = (os.geteuid() == 0)  # type: ignore[attr-defined]  # available in Linux
    is_root_user = (os.getuid() == 0)  # type: ignore[attr-defined]  # available in Linux
    if running_as_root_user and not is_root_user:
        print('*** --install-to-desktop should not be run as root or with sudo')
        sys.exit(1)
    
    # Install .desktop file to ~/.local/share/applications
    # 
    # NOTE: Only .desktop files opened from this directory will show their
    #       icon in the dock correctly.
    if True:
        # Format .desktop file in memory
        with resources.open_text('crystal.desktop', encoding='utf-8') as f:
            desktop_file_content = f.read()
        if os.path.basename(sys.executable) in ['python', 'python3']:
            crystal_executable = f'{sys.executable} -m crystal'
        else:
            crystal_executable = sys.executable
        desktop_file_content = desktop_file_content.replace(
            '__CRYSTAL_PATH__', crystal_executable)
        desktop_file_content = desktop_file_content.replace(
            '__APPICON_PATH__', resources.get_filepath('appicon.png'))
        
        apps_dirpath = os.path.expanduser('~/.local/share/applications')
        os.makedirs(apps_dirpath, exist_ok=True)
        with open(f'{apps_dirpath}/crystal.desktop', 'w') as f:
            f.write(desktop_file_content)
        
        subprocess.run([
            'update-desktop-database',
            os.path.expanduser('~/.local/share/applications')
        ], check=True)
    
    # Install symlink to .desktop file on ~/Desktop, if possible
    desktop_dirpath = os.path.expanduser('~/Desktop')
    if os.path.isdir(desktop_dirpath) and not os.path.exists(f'{desktop_dirpath}/crystal.desktop'):
        subprocess.run([
            'ln', '-s',
            f'{apps_dirpath}/crystal.desktop',
            f'{desktop_dirpath}/crystal.desktop',
        ], check=True)
        
        # Mark .desktop symlink on desktop as "Allow Launching" on Ubuntu 22+
        # https://askubuntu.com/questions/1218954/desktop-files-allow-launching-set-this-via-cli
        try:
            gio.set(f'{desktop_dirpath}/crystal.desktop', 'metadata::trusted', 'true')
        except gio.GioNotAvailable:
            pass
        except gio.UnrecognizedGioAttributeError:
            # For example Kubuntu 22 does not recognize this attribute
            pass
        else:
            subprocess.run([
                'chmod', 'a+x',
                f'{desktop_dirpath}/crystal.desktop',
            ], check=True)
    
    # Install .crystalopen MIME type definition
    mime_dirpath = os.path.expanduser('~/.local/share/mime')
    os.makedirs(f'{mime_dirpath}/packages', exist_ok=True)
    with open(f'{mime_dirpath}/packages/application-vnd.crystal.opener.xml', 'w') as dst_file:
        with resources.open_text('application-vnd.crystal.opener.xml', encoding='utf-8') as src_file:
            shutil.copyfileobj(src_file, dst_file)
    subprocess.run(['update-mime-database', mime_dirpath], check=True)
    
    # Install .crystalopen file icon
    # 
    # NOTE: It is necessary to install both PNG and SVG versions,
    #       because some desktop environments (like KDE) only recognize one type.
    # NOTE: It is necessary to install prescaled icon versions,
    #       because some apps (like GNOME/KDE's Open Dialog) cannot scale icons themselves.
    if True:
        # Locate places where MIME icons are installed, in global theme directories
        dimension_for_mime_icon_dirpath = {}  # type: Dict[str, Union[Tuple[int, int], EllipsisType]]
        with _cwd_set_to('/usr/share/icons'):
            for mime_icon_filepath in glob.iglob('**/text-plain.*', recursive=True):
                mime_icon_dirpath = os.path.dirname(mime_icon_filepath)
                
                old_dimension = dimension_for_mime_icon_dirpath.get(mime_icon_dirpath)
                if old_dimension is not Ellipsis:
                    with open(mime_icon_filepath, 'rb') as f:
                        if mime_icon_filepath.endswith('.png'):
                            new_dimension = get_png_dimensions(f) or Ellipsis
                        elif mime_icon_filepath.endswith('.svg'):
                            new_dimension = get_svg_dimensions(f) or Ellipsis
                        else:
                            # Ignore futuristic image types
                            continue
                    if old_dimension is not None and new_dimension != old_dimension:
                        new_dimension = Ellipsis  # ambiguous
                    if new_dimension != old_dimension:
                        dimension_for_mime_icon_dirpath[mime_icon_dirpath] = new_dimension
        
        # Install new MIME icons, in local theme directories
        if len(dimension_for_mime_icon_dirpath) == 0:
            print('WARNING: Unable to locate places to install MIME icons')
        else:
            local_icons_dirpath = os.path.expanduser('~/.local/share/icons')
            for (mime_icon_dirpath, dimension) in dimension_for_mime_icon_dirpath.items():
                mime_icon_abs_dirpath = os.path.join(local_icons_dirpath, mime_icon_dirpath)
                os.makedirs(mime_icon_abs_dirpath, exist_ok=True)
                
                # Install PNG icon for MIME type
                with open(f'{mime_icon_abs_dirpath}/application-vnd.crystal.opener.png', 'wb') as dst_file:
                    dst_file.write(_get_or_load_best_icon('application-vnd.crystal.opener', 'png', dimension))
                
                # Install SVG icon for MIME type,
                # because at least KDE on Kubuntu 22 seems to ignore PNG icons
                with open(f'{mime_icon_abs_dirpath}/application-vnd.crystal.opener.svg', 'wb') as dst_file:
                    dst_file.write(_get_or_load_best_icon('application-vnd.crystal.opener', 'svg', dimension))
            
            # NOTE: At least on Ubuntu 22 it seems the icon caches don't need
            #       to be explicitly updated to pick up the new icon type
            #       immediately, which is good because it may not be possible
            #       to sudo.
            #subprocess.run(['sudo', 'update-icon-caches', *glob.glob('/usr/share/icons/*')], check=True)
    
    # Install .crystalproj folder icon
    # 
    # NOTE: It is necessary to install both PNG and SVG versions,
    #       because some desktop environments (like KDE) only recognize one type.
    # NOTE: It is necessary to install prescaled icon versions,
    #       because some apps (like GNOME/KDE's Open Dialog) cannot scale icons themselves.
    if True:
        # Determine icon names for regular folder, which actually resolve to an icon
        icon_names: list[str]
        try:
            regular_folder_dirpath = os.path.dirname(__file__)
            p = subprocess.run(
                ['gio', 'info', '--attributes', 'standard::icon', regular_folder_dirpath],
                check=True,
                capture_output=True,
            )
        except FileNotFoundError:
            # GIO not available
            icon_names = []
        else:
            gio_lines = p.stdout.decode('utf-8').split('\n')
            for line in gio_lines:
                line = line.strip()  # reinterpret
                if line.startswith('standard::icon:'):
                    line = line[len('standard::icon:'):]  # reinterpret
                    icon_names = [x.strip() for x in line.strip().split(',')]
                    break
            else:
                icon_names = []
            
        # Locate places where folder icons are installed, in global theme directories
        dimension_for_folder_icon_dirpath = {}  # type: Dict[str, Union[Tuple[int, int], EllipsisType]]
        if len(icon_names) != 0:
            with _cwd_set_to('/usr/share/icons'):
                for (folder_icon_dirpath, dirnames, filenames) in os.walk('.'):
                    for filename in filenames:
                        (filename_without_ext, _) = os.path.splitext(filename)
                        if filename_without_ext not in icon_names:
                            continue
                        
                        old_dimension = dimension_for_folder_icon_dirpath.get(folder_icon_dirpath)
                        if old_dimension is not Ellipsis:
                            with open(os.path.join(folder_icon_dirpath, filename), 'rb') as f:
                                if filename.endswith('.png'):
                                    new_dimension = get_png_dimensions(f) or Ellipsis
                                elif filename.endswith('.svg'):
                                    new_dimension = get_svg_dimensions(f) or Ellipsis
                                else:
                                    # Ignore futuristic image types
                                    continue
                            if old_dimension is not None and new_dimension != old_dimension:
                                new_dimension = Ellipsis  # ambiguous
                            if new_dimension != old_dimension:
                                dimension_for_folder_icon_dirpath[folder_icon_dirpath] = new_dimension
        
        # Install new folder icons, in local theme directories
        if len(dimension_for_folder_icon_dirpath) == 0:
            print('WARNING: Unable to locate places to install folder icons')
        else:
            local_icons_dirpath = os.path.expanduser('~/.local/share/icons')
            for (folder_icon_dirpath, dimension) in dimension_for_folder_icon_dirpath.items():
                folder_icon_abs_dirpath = os.path.join(local_icons_dirpath, folder_icon_dirpath)
                os.makedirs(folder_icon_abs_dirpath, exist_ok=True)
                
                # Install PNG icon for folder type
                with open(f'{folder_icon_abs_dirpath}/crystalproj.png', 'wb') as dst_file:
                    dst_file.write(_get_or_load_best_icon('docicon', 'png', dimension))
                
                # Install SVG icon for folder type,
                # because at least KDE on Kubuntu 22 seems to ignore PNG icons
                with open(f'{folder_icon_abs_dirpath}/crystalproj.svg', 'wb') as dst_file:
                    dst_file.write(_get_or_load_best_icon('docicon', 'svg', dimension))
    
    # If KDE, restart plasmasession so that desktop detects new icons immediately
    import psutil  # type: ignore[reportMissingModuleSource]  # Linux-only dependency
    for process in psutil.process_iter(attrs=['cmdline', 'pid', 'cwd', 'environ']):
        process_cmdline = process.cmdline()
        if (len(process_cmdline) >= 1 and
                os.path.basename(process_cmdline[0]) == 'plasmashell'):
            plasmashell = process  # type: Optional[psutil.Process]
            break
    else:
        plasmashell = None
    if plasmashell is not None:
        plasmashell_info = dict(
            cmdline=plasmashell.cmdline(),
            cwd=plasmashell.cwd(),
            environ=plasmashell.environ(),
        )  # capture
        subprocess.run(
            ['kill', str(plasmashell.pid)],
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        subprocess.Popen(
            plasmashell_info['cmdline'],
            cwd=plasmashell_info['cwd'],
            env=plasmashell_info['environ'],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        # (Leave new plasmashell process running)


@cache  # ...in memory
def _get_or_load_best_icon(
        name: str,
        type: Literal['png', 'svg'],
        dimension: tuple[int, int] | EllipsisType,
        ) -> bytes:
    """
    Returns the bytes for the icon with the specified name and type
    which either has dimensions matching the specified dimensions or
    is the fallback version of the icon.
    """
    from crystal import resources
    
    if dimension == Ellipsis:
        icon_filename = f'{name}.{type}'
    else:
        assert not isinstance(dimension, EllipsisType)  # help mypy
        icon_filename = f'{name}_{dimension[0]}x{dimension[1]}.{type}'
    try:
        with resources.open_binary(icon_filename) as src_file:
            return src_file.read()
    except FileNotFoundError:
        if dimension == Ellipsis:
            raise ValueError(f'Fallback icon {icon_filename!r} not found')
        return _get_or_load_best_icon(name, type, Ellipsis)


# ------------------------------------------------------------------------------
# Utility: Read Image Dimensions

_PNG_SIGNATURE = bytes([137, 80, 78, 71, 13, 10, 26, 10])  # b'\x89PNG\r\n\x1a\n'


def get_png_dimensions(f: BinaryIO) -> tuple[int, int] | None:
    """
    Reads the specified PNG file object to determine its (width, height) dimensions.
    Returns None if the file is not a valid PNG file.
    """
    # PNG specification 1.2:
    # http://libpng.org/pub/png/spec/1.2/PNG-Contents.html
    
    # Read/validate PNG file signature
    f.seek(0)
    signature = f.read(len(_PNG_SIGNATURE))
    if signature != _PNG_SIGNATURE:
        # Not a PNG file
        return None
    
    # Read IHDR chunk, containing width and height dimensions
    # 
    # NOTE: The CRC of the IHDR chunk is NOT validated for simplicity of implementation
    f.seek(12)
    chunk_type = f.read(4)
    if chunk_type != b'IHDR':
        # PNG file missing required first 'IHDR' chunk
        return None
    width = _decode_uint(f.read(4))
    height = _decode_uint(f.read(4))
    if width == 0 or height == 0:
        # Invalid width or height
        return None
    return (width, height)


def get_svg_dimensions(f: BinaryIO) -> tuple[int, int] | None:
    """
    Reads the specified SVG file object to determine its (width, height) dimensions.
    Returns None if the file is not a valid SVG file or no dimensions are defined.
    """
    try:
        for (event, el) in ET.iterparse(f, events=('start',)):
            assert event == 'start'
            first_el = el
            
            # Ensure initial tag is <svg>
            if first_el.tag != '{http://www.w3.org/2000/svg}svg':
                # Not an SVG file
                return None
            
            # Parse "viewBox" from <svg> element
            view_box_str = first_el.attrib.get('viewBox')
            if view_box_str is None:
                # No view box defined. No dimensions available.
                return None
            try:
                view_box_int_parts = [int(str_part) for str_part in view_box_str.split(' ')]
            except ValueError:
                # Malformed "viewBox" attribute
                return None
            if len(view_box_int_parts) != 4:
                # Malformed "viewBox" attribute
                return None
            (_, _, width, height) = view_box_int_parts
            if width > 0 and height > 0:
                return (width, height)
            else:
                # SVG rendering is disabled
                return None
        
        # Not an XML file (or an SVG file)
        return None
    except ET.ParseError:
        # Not an XML file (or an SVG file)
        return None


def _decode_uint(uint_bytes: bytes) -> int:
    value = 0
    for b in uint_bytes:
        value = (value << 8) + b
    return value


# ------------------------------------------------------------------------------
# Utility: Change Working Directory

@contextmanager
def _cwd_set_to(dirpath: str) -> 'Iterator[None]':
    old_cwd = os.getcwd()
    os.chdir(dirpath)
    try:
        yield
    finally:
        os.chdir(old_cwd)


# ------------------------------------------------------------------------------



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from crystal import resources
from crystal.model import (
    Project, Resource, ResourceGroup, ResourceRevision, RootResource,
)
from crystal.server import ProjectServer
from crystal.tests.util.controls import click_button, TreeItem
from crystal.tests.util.runner import bg_fetch_url
from crystal.tests.util.server import (
    assert_does_open_webbrowser_to, extracted_project,
    served_project_from_filepath,
)
from crystal.tests.util.tasks import wait_for_download_to_start_and_finish
from crystal.tests.util.wait import (
    first_child_of_tree_item_is_not_loading_condition, wait_for,
)
from crystal.tests.util.windows import (
    EntityTree, MainWindow, OpenOrCreateDialog,
)
import os
import tempfile
from unittest import skip
from unittest.mock import ANY

# All of the following tests have the implicit prefix:
# - test_given_html_page_links_to_embedded_resource_in_a_do_not_download_group...


async def test_when_download_html_page_then_does_not_download_embedded_resource_automatically() -> None:
    async with _project_with_do_not_download_group_open() as (
            mw, project, home_url, comic_image_rg_pattern, comic_image_r_url):
        # Download HTML page
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        home_ti = root_ti.GetFirstChild()
        assert home_ti is not None
        home_ti.Expand()
        await wait_for_download_to_start_and_finish(mw.task_tree)
        assert first_child_of_tree_item_is_not_loading_condition(home_ti)()
        
        # Expand HTML page children in entity tree
        comic_image_rg_ti = home_ti.find_child(comic_image_rg_pattern)
        comic_image_rg_ti.Expand()
        await wait_for(first_child_of_tree_item_is_not_loading_condition(comic_image_rg_ti))
        comic_image_rg_ti.find_child(comic_image_r_url)
        
        home_r = project.get_resource(home_url)
        assert home_r is not None
        assert home_r.has_any_revisions()
        
        comic_image_r = project.get_resource(comic_image_r_url)
        assert comic_image_r is not None
        assert not comic_image_r.has_any_revisions()


async def test_then_embedded_resource_does_not_appear_in_a_hidden_embedded_cluster_in_entity_tree() -> None:
    async with _project_with_do_not_download_group_open() as (
            mw, project, home_url, comic_image_rg_pattern, comic_image_r_url):
        # Download HTML page
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        home_ti = root_ti.GetFirstChild()
        assert home_ti is not None
        home_ti.Expand()
        await wait_for_download_to_start_and_finish(mw.task_tree)
        assert first_child_of_tree_item_is_not_loading_condition(home_ti)()
        
        # Expand HTML page children in entity tree
        (hidden_embedded_ti,) = (
            child for child in home_ti.Children
            if child.Text == '(Hidden: Embedded)'
        )
        hidden_embedded_ti.Expand()
        await wait_for(first_child_of_tree_item_is_not_loading_condition(hidden_embedded_ti))
        
        assert None == hidden_embedded_ti.try_find_child(comic_image_r_url)


async def test_when_browse_to_html_page_and_browser_requests_embedded_resource_then_do_not_dynamically_download_embedded_resource_and_instead_return_http_404() -> None:
    async with _project_with_do_not_download_group_open() as (
            mw, project, home_url, comic_image_rg_pattern, comic_image_r_url):
        # Download HTML page
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        home_ti = root_ti.GetFirstChild()
        assert home_ti is not None
        home_ti.SelectItem()
        await mw.click_download_button()
        await wait_for_download_to_start_and_finish(mw.task_tree)
        
        # Start server
        home_ti.SelectItem()
        with assert_does_open_webbrowser_to(ANY):
            click_button(mw.view_button)
        server = ProjectServer._last_created
        assert server is not None
        
        response = await bg_fetch_url(server.get_request_url(comic_image_r_url))
        assert 404 == response.status


async def test_given_embedded_resource_selected_in_entity_tree_when_press_download_button_explicitly_then_downloads_embedded_resource() -> None:
    async with _project_with_do_not_download_group_open() as (
            mw, project, home_url, comic_image_rg_pattern, comic_image_r_url):
        # Download HTML page
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        home_ti = root_ti.GetFirstChild()
        assert home_ti is not None
        home_ti.Expand()
        await wait_for_download_to_start_and_finish(mw.task_tree)
        assert first_child_of_tree_item_is_not_loading_condition(home_ti)()
        
        # Expand HTML page children in entity tree
        comic_image_rg_ti = home_ti.find_child(comic_image_rg_pattern)
        comic_image_rg_ti.Expand()
        await wait_for(first_child_of_tree_item_is_not_loading_condition(comic_image_rg_ti))
        comic_image_r_ti = comic_image_rg_ti.find_child(comic_image_r_url)
        
        comic_image_r = project.get_resource(comic_image_r_url)
        assert comic_image_r is not None
        assert not comic_image_r.has_any_revisions()
        
        comic_image_r_ti.SelectItem()
        await mw.click_download_button(
            # NOTE: May "finish immediately" because has no embedded subresources
            immediate_finish_ok=True)
        await wait_for_download_to_start_and_finish(
            mw.task_tree,
            # NOTE: May "finish immediately" because has no embedded subresources
            immediate_finish_ok=True)
        assert comic_image_r.has_any_revisions()


async def test_given_do_not_download_group_selected_in_entity_tree_when_press_download_button_explicitly_then_downloads_group() -> None:
    async with _project_with_do_not_download_group_open() as (
            mw, project, home_url, comic_image_rg_pattern, comic_image_r_url):
        # Download HTML page
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        home_ti = root_ti.GetFirstChild()
        assert home_ti is not None
        home_ti.SelectItem()
        await mw.click_download_button()
        await wait_for_download_to_start_and_finish(mw.task_tree)
        
        # Expand group children in entity tree
        comic_image_rg_ti = root_ti.find_child(comic_image_rg_pattern)
        comic_image_rg_ti.Expand()
        await wait_for(first_child_of_tree_item_is_not_loading_condition(comic_image_rg_ti))
        comic_image_rg_ti.find_child(comic_image_r_url)
        
        comic_image_r = project.get_resource(comic_image_r_url)
        assert comic_image_r is not None
        assert not comic_image_r.has_any_revisions()
        
        comic_image_rg_ti.SelectItem()
        await mw.click_download_button()
        await wait_for_download_to_start_and_finish(mw.task_tree)
        assert comic_image_r.has_any_revisions()


async def test_then_embedded_resource_in_entity_tree_appears_with_do_not_download_badge() -> None:
    async with _project_with_do_not_download_group_open() as (
            mw, project, home_url, comic_image_rg_pattern, comic_image_r_url):
        # Download HTML page
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        home_ti = root_ti.GetFirstChild()
        assert home_ti is not None
        home_ti.Expand()
        await wait_for_download_to_start_and_finish(mw.task_tree)
        assert first_child_of_tree_item_is_not_loading_condition(home_ti)()
        
        # Expand HTML page children in entity tree
        comic_image_rg_ti = home_ti.find_child(comic_image_rg_pattern)
        comic_image_rg_ti.Expand()
        await wait_for(first_child_of_tree_item_is_not_loading_condition(comic_image_rg_ti))
        comic_image_r_ti = comic_image_rg_ti.find_child(comic_image_r_url)
        
        await _assert_tree_item_icon_tooltip_contains(comic_image_rg_ti, 'Ignored')
        await _assert_tree_item_icon_tooltip_contains(comic_image_r_ti, 'Ignored')
        
        # Expand group children in entity tree
        comic_image_rg_ti = root_ti.find_child(comic_image_rg_pattern)
        comic_image_rg_ti.Expand()
        await wait_for(first_child_of_tree_item_is_not_loading_condition(comic_image_rg_ti))
        comic_image_r_ti = comic_image_rg_ti.find_child(comic_image_r_url)
        
        await _assert_tree_item_icon_tooltip_contains(comic_image_rg_ti, 'Ignored')
        await _assert_tree_item_icon_tooltip_contains(comic_image_r_ti, 'Ignored')
        
        # Test: test_given_embedded_resource_also_in_a_non_do_not_download_group_then_embedded_resource_in_entity_tree_does_not_appear_with_do_not_download_badge
        if True:
            specific_comic_image_rg = ResourceGroup(
                project, 'air_gap_2x.png', comic_image_r_url)
            await _assert_tree_item_icon_tooltip_contains(comic_image_r_ti, 'Undownloaded')
            
            specific_comic_image_rg.delete()
            await _assert_tree_item_icon_tooltip_contains(comic_image_r_ti, 'Ignored')


@skip('covered by: test_then_embedded_resource_in_entity_tree_appears_with_do_not_download_badge')
async def test_then_do_not_download_group_in_entity_tree_appears_with_do_not_download_badge() -> None:
    pass


@skip('covered by: test_then_embedded_resource_in_entity_tree_appears_with_do_not_download_badge')
async def test_given_embedded_resource_also_in_a_non_do_not_download_group_then_embedded_resource_in_entity_tree_does_not_appear_with_do_not_download_badge() -> None:
    pass


async def test_when_reopen_project_then_group_that_was_marked_as_do_not_download_is_still_marked_as_do_not_download() -> None:
    # Define URLs
    comic_image_rg_pattern = 'https://imgs.xkcd.com/comics/*'
    
    # Case 1: New group marked as do_not_download
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        os.rmdir(project_dirpath)
        
        # Create project with do_not_download group
        with Project(project_dirpath) as project:
            comic_image_rg1 = ResourceGroup(
                project, 'Comic Image', comic_image_rg_pattern,
                do_not_download=True)
            assert True == comic_image_rg1.do_not_download
        
        # Reopen project. Ensure group is still marked as do_not_download
        with Project(project_dirpath) as project:
            comic_image_rg2 = project.get_resource_group(name='Comic Image')
            assert comic_image_rg2 is not None
            assert True == comic_image_rg2.do_not_download
    
    # Case 2: Edited group marked as do_not_download
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        os.rmdir(project_dirpath)
        
        # Create project with group. Edit to mark as do_not_download.
        with Project(project_dirpath) as project:
            comic_image_rg1 = ResourceGroup(
                project, 'Comic Image', comic_image_rg_pattern,
                do_not_download=False)
            assert False == comic_image_rg1.do_not_download
            
            comic_image_rg1.do_not_download = True
            assert True == comic_image_rg1.do_not_download
        
        # Reopen project. Ensure group is still marked as do_not_download
        with Project(project_dirpath) as project:
            comic_image_rg2 = project.get_resource_group(name='Comic Image')
            assert comic_image_rg2 is not None
            assert True == comic_image_rg2.do_not_download


# ------------------------------------------------------------------------------
# Utility

@asynccontextmanager
async def _project_with_do_not_download_group_open(
        ) -> AsyncIterator[tuple[MainWindow, Project, str, str, str]]:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Add comic image to project,
        # since it's not part of the default project test data
        with Project(project_dirpath) as project1:
            r = Resource(project1, 'https://imgs.xkcd.com/comics/air_gap_2x.png')
            with resources.open_binary('testdata_air_gap_2x.png') as image_stream:
                rr = ResourceRevision.create_from_response(
                    r,
                    metadata=None,
                    body_stream=image_stream)
        
        with served_project_from_filepath(project_dirpath) as sp:
            # Define URLs
            home_url = sp.get_request_url('https://xkcd.com/')
            comic_image_rg_pattern = sp.get_request_url('https://imgs.xkcd.com/comics/*')
            comic_image_r_url = sp.get_request_url('https://imgs.xkcd.com/comics/air_gap_2x.png')
            
            async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
                # Define entities
                RootResource(project, 'Home', Resource(project, home_url))
                comic_image_rg = ResourceGroup(
                    project, 'Comic Image', comic_image_rg_pattern,
                    do_not_download=True)
                
                yield (mw, project, home_url, comic_image_rg_pattern, comic_image_r_url)


# NOTE: Only for use with tree items in EntityTree
_assert_tree_item_icon_tooltip_contains = EntityTree.assert_tree_item_icon_tooltip_contains


# ------------------------------------------------------------------------------



from __future__ import annotations

from collections.abc import Callable
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.screenshots import take_error_screenshot
import datetime
import os
import time
from typing import Literal, Optional, TYPE_CHECKING, TypeVar
import warnings
import wx

if TYPE_CHECKING:
    from crystal.tests.util.controls import TreeItem


_T = TypeVar('_T')

_T1 = TypeVar('_T1')
_T2 = TypeVar('_T2')
_T3 = TypeVar('_T3')


GLOBAL_TIMEOUT_MULTIPLIER = (
    float(os.environ.get('CRYSTAL_GLOBAL_TIMEOUT_MULTIPLIER', '1.0'))
)


# The hard timeout for waits = the soft timeout * HARD_TIMEOUT_MULTIPLIER
HARD_TIMEOUT_MULTIPLIER = 2.0


# ------------------------------------------------------------------------------
# Utility: Wait While

async def wait_while(
        progression_func: Callable[[], _T | None],
        *, progress_timeout: float | None=None,
        progress_timeout_message: Callable[[], str] | None=None,
        period: float | None=None,
        ) -> None:
    """
    Waits while the specified progression returns different non-None values
    at least every `progress_timeout` seconds, checking every `period` seconds,
    until a None value is returned.
    
    Raises:
    * WaitTimedOut -- 
        if the `progress_timeout` expires while waiting for
        a differing value from the specified progression
    """
    def print_new_status(status: object) -> None:
        print('[' + datetime.datetime.now().strftime('%H:%M:%S.%f') + ']' + ' ' + str(status))
    
    last_status = progression_func()
    if last_status is None:
        return  # done
    print_new_status(last_status)
    
    def do_check_status() -> bool | None:
        nonlocal last_status
        
        current_status = progression_func()
        if current_status is None:
            print_new_status('DONE')
            return True  # done
        
        changed_status = (current_status != last_status)  # capture
        last_status = current_status  # reinterpret
        
        if changed_status:
            print_new_status(current_status)
            return False  # progress
        else:
            return None  # no progress
    
    while True:
        is_done = (await wait_for(
            do_check_status,
            timeout=progress_timeout,
            period=period,
            message=progress_timeout_message,
        ))  # type: bool
        if is_done:
            return


# ------------------------------------------------------------------------------
# Utility: Wait For

DEFAULT_WAIT_TIMEOUT = 2.0  # arbitrary
DEFAULT_WAIT_PERIOD = 0.1  # arbitrary


async def wait_for(
        condition: Callable[[], _T | None],
        timeout: float | None=None,
        *, period: float | None=None,
        message: Callable[[], str] | None=None,
        stacklevel_extra: int=0,
        screenshot_on_error: bool=True,
        ) -> _T:
    """
    Waits up to `timeout` seconds for the specified condition to become non-None,
    returning the result of the condition, checking every `period` seconds.
    
    The condition is always checked on the foreground thread.
    The foreground thread is released while waiting between checks.
    
    Raises:
    * WaitTimedOut -- if the timeout expires before the condition becomes non-None
    """
    if timeout is None:
        timeout = DEFAULT_WAIT_TIMEOUT
    if period is None:
        period = DEFAULT_WAIT_PERIOD
    
    timeout *= GLOBAL_TIMEOUT_MULTIPLIER  # reinterpret
    
    soft_timeout = timeout
    hard_timeout = timeout * HARD_TIMEOUT_MULTIPLIER
    
    start_time = time.time()  # capture
    hard_timeout_exceeded = False
    try:
        while True:
            condition_result = condition()
            if condition_result is not None:
                return condition_result
            
            # Raise if hard timeout exceeded
            delta_time = time.time() - start_time
            if delta_time > hard_timeout:
                if message is not None:
                    # Use caller-provided failure message if available
                    message_str = message()
                elif hasattr(condition, 'description'):
                    condition_description = condition.description  # type: ignore[attr-defined]
                    message_str = f'Timed out waiting {timeout}s for {condition_description}'
                else:
                    message_str = f'Timed out waiting {timeout}s for {condition!r}'
                
                # Screenshot the timeout error
                if screenshot_on_error:
                    take_error_screenshot()
                
                hard_timeout_exceeded = True
                raise WaitTimedOut(message_str)
            
            await bg_sleep(period)
    finally:
        # Warn if soft timeout exceeded
        if not hard_timeout_exceeded:
            delta_time = time.time() - start_time
            if delta_time > soft_timeout:
                message_suffix_str = None
                if message is not None:
                    # Use caller-provided failure message if available
                    message_suffix_str = message()
                elif hasattr(condition, 'description'):
                    condition_description = condition.description  # type: ignore[attr-defined]
                    message_suffix_str = f'{condition_description}'
                else:
                    message_suffix_str = f'{condition!r}'
                
                warnings.warn(
                    'Soft timeout exceeded ({:.1f}s > {:.1f}s). {}'.format(
                        delta_time,
                        soft_timeout,
                        message_suffix_str
                    ),
                    stacklevel=(2 + stacklevel_extra))


def wait_for_sync(condition: Callable[[], _T | None], *args, **kwargs) -> _T:
    """
    Similar to wait_for() but does not release the current thread between waits.
    """
    from crystal.tests.util.runner import SleepCommand
    coro = wait_for(condition, *args, **kwargs)
    while True:
        try:
            command = coro.send(None)
        except StopIteration as e:
            return e.value
        assert isinstance(command, SleepCommand)
        time.sleep(command.delay)


class WaitTimedOut(Exception):
    pass


def window_condition(
        name: str, *, hidden_ok: bool=False
        ) -> Callable[[], wx.Window | None]:
    """
    Whether the named window exists and is visible.
    Truthy return value is the window.
    """
    def window() -> wx.Window | None:
        window = wx.FindWindowByName(name)  # type: Optional[wx.Window]
        if window is None:
            return None
        if not hidden_ok and not window.IsShown():
            return None
        return window
    window.description = (  # type: ignore[attr-defined]
        f'window {name!r} to appear'
        if not hidden_ok
        else f'window {name!r} to be created'
    )
    return window


def first_child_of_tree_item_is_not_loading_condition(
        ti: TreeItem
        ) -> Callable[[], wx.TreeItemId | None]:
    """
    Whether the specified tree item's children is done loading.
    Truthy return value is the first loaded child.
    """
    def first_child_of_tree_item_is_not_loading() -> TreeItem | None:
        first_child_ti = ti.GetFirstChild()
        if first_child_ti is None:
            return None
        if first_child_ti.Text == 'Loading...':
            return None
        return first_child_ti
    return first_child_of_tree_item_is_not_loading


def tree_has_children_condition(
        tree: wx.TreeCtrl,
        ) -> Callable[[], Literal[True] | None]:
    """Whether the specified tree has children."""
    return not_condition(tree_has_no_children_condition(tree))


def tree_has_no_children_condition(
        tree: wx.TreeCtrl, 
        ) -> Callable[[], Literal[True] | None]:
    """Whether the specified tree has no children."""
    from crystal.tests.util.controls import TreeItem
    return tree_item_has_no_children_condition(TreeItem(tree, tree.GetRootItem()))


def tree_item_has_no_children_condition(
        ti: TreeItem
        ) -> Callable[[], Literal[True] | None]:
    """Whether the specified tree item has no children."""
    def tree_item_has_no_children() -> Literal[True] | None:
        first_child_tii = ti.tree.GetFirstChild(ti.id)[0]
        if not first_child_tii.IsOk():
            return True
        else:
            return None
    return tree_item_has_no_children


def is_enabled_condition(window: wx.Window) -> Callable[[], Literal[True] | None]:
    """Whether the specified window is enabled."""
    def is_enabled() -> Literal[True] | None:
        return window.Enabled or None
    return is_enabled


def not_condition(
        condition: Callable[[], _T | None]
        ) -> Callable[[], Literal[True] | None]:
    """Whether the specified condition is falsy."""
    def not_() -> Literal[True] | None:
        if condition():
            return None
        else:
            return True
    return not_


def or_condition(
        *conditions: Callable[[], _T | None]
        ) -> Callable[[], _T | None]:
    """
    Whether any of the specified conditions are true.
    Truthy return value is the return value of the first truthy condition.
    """
    def or_() -> _T | None:
        for condition in conditions:
            result = condition()
            if result is not None:
                return result
        return None
    return or_


# ------------------------------------------------------------------------------



from collections.abc import Iterator
from contextlib import contextmanager
import gc
import os
import sys
import time

# If True, then the runtime of foreground tasks is tracked to ensure
# they are short. This is necessary to keep the UI responsive.
PROFILE_GC = os.environ.get('CRYSTAL_NO_PROFILE_GC', 'False') != 'True'

# Maximum reasonable time that foreground tasks should take to complete.
# If profiling is enabled, warnings will be printed for tasks whose runtime
# exceeds this threshold.
_GC_RUNTIME_THRESHOLD = 1.0 # sec


def start_profiling_gc() -> None:
    last_gc_start = None
    def on_gc(phase: str, info: dict) -> None:
        nonlocal last_gc_start
        now = time.time()  # capture
        if phase == 'start':
            last_gc_start = now
        elif phase == 'stop':
            if last_gc_start is not None:
                duration = now - last_gc_start
                if duration > _GC_RUNTIME_THRESHOLD:
                    print('*** {} took {:.02f}s to execute: {!r}'.format(
                        'Garbage collection',
                        duration,
                        info,
                    ), file=sys.stderr)
                
                last_gc_start = None
    gc.callbacks.append(on_gc)


@contextmanager
def gc_disabled() -> Iterator[None]:
    """
    Context in which garbage collection is disabled.
    
    Useful when allocating a very large number of long-lived objects,
    to avoid many useless runs of the garbage collector.
    """
    was_enabled = gc.isenabled()
    if was_enabled:
        gc.disable()
    try:
        yield
    finally:
        assert not gc.isenabled()
        if was_enabled:
            gc.enable()



from base64 import urlsafe_b64decode, urlsafe_b64encode
from crystal.plugins.util.params import try_get_int, try_get_str
import json
import sys
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse


def normalize_url(old_url: str, **kwargs) -> str:
    old_url_parts = urlparse(old_url)
    if old_url_parts.scheme in ('http', 'https'):
        # https?://__DOMAIN__/api/v1/firehose?_=...&d=...
        if old_url_parts.path == '/api/v1/firehose':
            params = parse_qs(old_url_parts.query)
            t = try_get_int(params, '_')
            d = try_get_str(params, 'd')
            if t is not None and d is not None:
                try:
                    d_obj = json.loads(urlsafe_b64decode(d))  # type: ignore[attr-defined]
                except ValueError:
                    print('*** Substack: Unable to decode "d" argument: ' + d, file=sys.stderr)
                else:
                    if isinstance(d_obj, dict):
                        # Pin $.properties.browserSessionId to fixed arbitrary valid value
                        properties = d_obj.get('properties', {})
                        browserSessionId = properties.get('browserSessionId', None)
                        if isinstance(browserSessionId, str):
                            browserSessionId = '1xh68gok0s4'  # reinterpret
                            properties['browserSessionId'] = browserSessionId  # reinterpret
                        
                        # Pin $.context.page.referrer to '' (which usually is already the case)
                        context = d_obj.get('context', {})
                        page = context.get('page', {})
                        referrer = page.get('referrer', None)
                        if isinstance(referrer, str):
                            referrer = ''  # reinterpret
                            page['referrer'] = referrer  # reinterpret
                        
                        # Alter $.context.page.url so that its domain matches that of old_url
                        url = page.get('url', None)
                        if isinstance(url, str):
                            url_parts = urlparse(url)
                            url = urlunparse(url_parts._replace(
                                scheme=old_url_parts.scheme,
                                netloc=old_url_parts.netloc,
                            ))  # reinterpret
                            page['url'] = url  # reinterpret
                        
                        new_d = urlsafe_b64encode(
                            json.dumps(d_obj).encode('utf-8')  # type: ignore[attr-defined]
                        ).decode('utf-8')
                        
                        new_params = dict(_=t, d=new_d)
                        new_query = urlencode(new_params)
                        return urlunparse(old_url_parts._replace(query=new_query))
    
    return old_url



from crystal.util.xos import is_windows
from tqdm import tqdm
from tqdm.std import EMA  # type: ignore[attr-defined]
from typing_extensions import override


class ProgressBarCalculator:
    """
    Calculates statistics about a process in progress,
    such as estimated time remaining.
    """
    _VERBOSE = False
    
    # Calibration notes:
    # - Minimalist Baker can take 10s - 17s to download a single recipe page.
    #   Even with a group growth rate of 0, that means a download rate between
    #   0.05 - 0.1.
    _MINIMUM_TOTAL_RATE_TO_REPORT_REMAINING_TIME = 0.0
    # Calibration notes:
    # - 0.10 seems "too low"
    # - 0.15 seems "too high"
    _MINIMUM_GROWTH_RATE_TO_REPORT = 0.13
    
    # Calibration notes:
    # - 10.0 seems "too high"
    _MAXIMUM_DELAY_BETWEEN_GROWTH_UPDATES = 5.0  # seconds
    
    # HACK: Avoid printing characters outside the Windows-1252 encoding (cp1252)
    #       when running on Windows because Windows seems to generally expect
    #       stdout/stderr on Windows to be decodable using that encoding.
    _DELTA = 'Δ' if not is_windows() else 'd'
    
    def __init__(self, initial: int, total: int) -> None:
        self._rc_n = RateCalculator(initial, total)
        self._rc_total = RateCalculator(total)
    
    # === Properties ===
    
    @property
    def n(self) -> int:
        return self._rc_n.n
    
    def _get_total(self) -> int:
        return self._rc_n.total
    def _set_total(self, total: int) -> None:
        delta_total = total - self._rc_n.total
        if delta_total:
            self._rc_n.total = total
            self._rc_total.update(delta_total)
    total = property(_get_total, _set_total)
    
    def remaining_str_and_time_per_item_str(self) -> tuple[str, str]:
        """Compute estimated remaining time, in format "MM:SS" or "HH:MM:SS"."""
        n = self.n
        total = self.total
        
        growth_rate_of_n = self._rc_n.rate
        growth_rate_of_total = self._rc_total.rate
        rate = (growth_rate_of_n or 0.0) - (growth_rate_of_total or 0.0)
        
        remaining = (
            (total - n) / rate
            if rate > self._MINIMUM_TOTAL_RATE_TO_REPORT_REMAINING_TIME and total
            else 0
        )
        # TODO: Call ProgressBarCalculator.format_interval() directly,
        #       rather than through a monkeypatch
        remaining_str = (
            tqdm.format_interval(remaining)
            if rate > self._MINIMUM_TOTAL_RATE_TO_REPORT_REMAINING_TIME
            else '?'
        )
        
        time_per_item_done = 1 / growth_rate_of_n if growth_rate_of_n else None
        time_per_item_add = 1 / growth_rate_of_total if growth_rate_of_total else None
        time_per_item_str = (
            f'{time_per_item_done:.2f}s'
            if time_per_item_done is not None
            else '?'
        ) + '/item'
        if (growth_rate_of_total or 0.0) > self._MINIMUM_GROWTH_RATE_TO_REPORT:
            assert time_per_item_add is not None
            time_per_item_str += ', ' + (
                f'{time_per_item_add:.2f}s'
            ) + '/growth'
        
        if self._VERBOSE:
            def format_floatlike(value: object) -> str:
                return (
                    f'{value:.2f}'
                    if isinstance(value, (float, int))
                    else f'{value}'
                )
            rate_str = (
                f'{rate:.2f}'
                if isinstance(rate, (float, int))
                else f'{rate}'
            )
            dn = format_floatlike(self._rc_n._tqdm._ema_dn())  # type: ignore[attr-defined]
            dt = format_floatlike(self._rc_n._tqdm._ema_dt())  # type: ignore[attr-defined]
            dntotal = format_floatlike(self._rc_total._tqdm._ema_dn())  # type: ignore[attr-defined]
            dttotal = format_floatlike(self._rc_total._tqdm._ema_dt())  # type: ignore[attr-defined]
            time_per_item_str += (
                f', rate={format_floatlike(rate)}, '
                f'{self._DELTA}n={dn}/'
                f'{dt}, '
                f'{self._DELTA}total={dntotal}/'
                f'{dttotal}'
            )
        
        return (remaining_str, time_per_item_str)
    
    # === Operations ===
    
    def update(self, delta_n: int) -> None:
        if not (delta_n >= 0):
            raise ValueError()
        self._rc_n._tqdm.update(delta_n)
        
        time_since_last_growth_update = (
            self._rc_n.last_print_t - self._rc_total.last_print_t
        )
        if time_since_last_growth_update > self._MAXIMUM_DELAY_BETWEEN_GROWTH_UPDATES:
            self._rc_total.update(0)
    
    def close(self) -> None:
        self._rc_n.close()
        self._rc_total.close()
    
    # === Utility ===
    
    @staticmethod
    def format_interval(t: float) -> str:
        """
        Formats a number of seconds as a clock time,
        `[H:]MM:SS` or `Dd + H:MM:SS`
        """
        mins, s = divmod(int(t), 60)
        hours, m = divmod(mins, 60)
        d, h = divmod(hours, 24)
        if d:
            # ex: '1d + 7:08:54'
            return f'{d:d}d + {h:d}:{m:02d}:{s:02d}'
        elif h:
            # ex: '7:08:54'
            return f'{h:d}:{m:02d}:{s:02d}'
        else:
            # ex: '08:54'
            return f'{m:02d}:{s:02d}'


class RateCalculator:
    def __init__(self, initial: int, total: int | None=None) -> None:
        self._tqdm = _TqdmWithoutDisplay(
            initial=initial,  # self.n = initial
            total=total,
            file=DevNullFile(),
            
            # Recompute remaining time after each update() call unless many
            # calls made in a short time interval (<= `mininterval`).
            miniters=0,
            mininterval=0.1,  # tqdm's default value
            
            # Calibrate EMA (exponential moving average) smoothing factors for speed estimates.
            # Ranges from 0.0 (average speed) to 1.0 (current/instantaneous speed).
            smoothing=0.1,  # NOTE: default is 0.3
        )
    
    # === Properties ===
    
    @property
    def n(self) -> int:
        return self._tqdm.n
    
    def _get_total(self) -> int | None:
        return self._tqdm.total
    def _set_total(self, total: int | None) -> None:
        self._tqdm.total = total
    total = property(_get_total, _set_total)
    
    @property
    def rate(self) -> float | None:
        dn = self._tqdm._ema_dn()  # type: ignore[attr-defined]  # cache
        dt = self._tqdm._ema_dt()  # type: ignore[attr-defined]  # cache
        rate = dn / dt if dt else None
        return rate
    
    @property
    def last_print_t(self) -> float:
        return self._tqdm.last_print_t
    
    # === Operations ===
    
    def update(self, delta_n: int) -> None:
        if not (delta_n >= 0):
            raise ValueError()
        if delta_n == 0:
            # Force the effect of: self._tqdm.update(0)
            cur_t = self._tqdm._time()  # type: ignore[attr-defined]
            dt = cur_t - self._tqdm.last_print_t
            dn = 0
            self._tqdm._ema_dn(dn)  # type: ignore[attr-defined]
            self._tqdm._ema_dt(dt)  # type: ignore[attr-defined]
            self._tqdm.last_print_t = cur_t
        else:
            self._tqdm.update(delta_n)
    
    def close(self) -> None:
        self._tqdm.close()


class _TqdmWithoutDisplay(tqdm):
    @override
    def refresh(self, nolock=False, lock_args=None):
        return super().refresh(nolock=True, lock_args=lock_args)
    
    @override
    def display(self, *args, **kwargs) -> None:
        pass  # do nothing


class DevNullFile:
    def write(self, value: str) -> None:
        pass
    
    def flush(self) -> None:
        pass


# ------------------------------------------------------------------------------

# Monkeypatch tqdm.format_interval() so that tqdm.format_meter() will use it
# when formatting timedeltas
tqdm.format_interval = ProgressBarCalculator.format_interval  # type: ignore[method-assign]



from __future__ import annotations

from collections.abc import AsyncIterator, Awaitable, Callable
from contextlib import AbstractContextManager, asynccontextmanager, nullcontext
from crystal.browser.new_root_url import (
    NewRootUrlDialog as RealNewRootUrlDialog,
)
from crystal.model import Project
from crystal.task import is_synced_with_scheduler_thread
from crystal.tests.util.controls import (
    click_button, file_dialog_returning, select_menuitem_now, TreeItem,
)
from crystal.tests.util.runner import bg_sleep, pump_wx_events
from crystal.tests.util.tasks import first_task_title_progression
from crystal.tests.util.wait import (
    not_condition, or_condition, tree_has_no_children_condition, wait_for,
    WaitTimedOut, window_condition,
)
from crystal.util.xos import is_mac_os
import os.path
import re
import sys
import tempfile
import traceback
from typing import TYPE_CHECKING
import wx

if TYPE_CHECKING:
    import wx.adv


# ------------------------------------------------------------------------------
# Window & Dialog Abstractions

class OpenOrCreateDialog:
    # NOTE: 10.0 isn't long enough for Windows test runners on GitHub Actions
    _TIMEOUT_FOR_OPEN_MAIN_WINDOW = 12.0
    
    open_or_create_project_dialog: wx.Dialog
    open_as_readonly: wx.CheckBox
    open_button: wx.Button
    create_button: wx.Button
    
    @staticmethod
    async def wait_for(timeout: float | None=None) -> OpenOrCreateDialog:
        self = OpenOrCreateDialog(ready=True)
        open_or_create_project_dialog = await wait_for(
            window_condition('cr-open-or-create-project'),
            timeout=timeout,
            stacklevel_extra=1,
        )  # type: wx.Window
        assert isinstance(open_or_create_project_dialog, wx.Dialog)
        self.open_or_create_project_dialog = open_or_create_project_dialog
        self.open_as_readonly = self.open_or_create_project_dialog.FindWindow(name=
            'cr-open-or-create-project__checkbox')
        assert isinstance(self.open_as_readonly, wx.CheckBox)
        self.open_button = self.open_or_create_project_dialog.FindWindow(id=wx.ID_NO)
        assert isinstance(self.open_button, wx.Button)
        self.create_button = self.open_or_create_project_dialog.FindWindow(id=wx.ID_YES)
        assert isinstance(self.create_button, wx.Button)
        return self
    
    def __init__(self, *, ready: bool=False) -> None:
        assert ready, 'Did you mean to use OpenOrCreateDialog.wait_for()?'
    
    @asynccontextmanager
    async def create(self, 
            project_dirpath: str | None=None,
            *, autoclose: bool=True,
            delete: bool | None=None,
            ) -> AsyncIterator[tuple[MainWindow, Project]]:
        """
        Creates a new project.
        
        If no project_dirpath is provided (i.e. the default) then a project
        will be created in a new temporary directory. That temporary directory
        will be deleted when the project is closed unless delete=False.
        
        If a project_dirpath is provided to an existing empty directory,
        a project will be created in that directory. That directory will
        not be deleted when the project is closed.
        
        The project will be automatically closed when exiting the
        returned context, unless autoclose=False, which leaves the
        MainWindow open.
        """
        if project_dirpath is None:
            if delete is None:
                delete = True
            
            # TODO: After upgrading to Python 3.12+, just use the
            #       "delete" parameter of TemporaryDirectory
            tmpdir_context = (
                tempfile.TemporaryDirectory(  # type: ignore[assignment]
                    suffix='.crystalproj',
                    # NOTE: If a file inside the temporary directory is still open,
                    #       ignore_cleanup_errors=True will prevent Windows from raising,
                    #       at the cost of leaving the temporary directory around
                    ignore_cleanup_errors=True
                ) if delete
                else nullcontext(tempfile.mkdtemp(suffix='.crystalproj'))
            )  # type: AbstractContextManager[str]
            
            with tmpdir_context as project_dirpath:
                assert project_dirpath is not None
                async with self.create(project_dirpath, autoclose=autoclose) as (mw, project2):
                    yield (mw, project2)
            return
        else:
            if delete is not None:
                raise ValueError(
                    'When creating a project in an existing directory, '
                    'the "delete" option should not be used.')
        
        mw = await self.create_and_leave_open(project_dirpath)
        
        project = Project._last_opened_project
        assert project is not None
        
        exc_info_while_close = None
        try:
            yield (mw, project)
        except:
            exc_info_while_close = sys.exc_info()
            raise
        finally:
            if autoclose:
                await mw.close(exc_info_while_close)
    
    async def create_and_leave_open(self, project_dirpath: str) -> MainWindow:
        old_opened_project = Project._last_opened_project  # capture
        try:
            with file_dialog_returning(project_dirpath):
                click_button(self.create_button)
                
                mw = await MainWindow.wait_for(timeout=self._TIMEOUT_FOR_OPEN_MAIN_WINDOW)
                return mw
        except:
            # Close any project that was opened
            new_opened_project = Project._last_opened_project  # capture
            if new_opened_project != old_opened_project:
                if new_opened_project is not None:
                    new_opened_project.close()
            
            raise
    
    @asynccontextmanager
    async def open(self, 
            project_dirpath: str, 
            *, readonly: bool | None=None,
            autoclose: bool=True,
            using_crystalopen: bool=False,
            wait_func: Callable[[], Awaitable[None]] | None=None,
            ) -> AsyncIterator[tuple[MainWindow, Project]]:
        """
        Opens an existing project.
        
        The project will be automatically closed when exiting the
        returned context, unless autoclose=False, which leaves the
        MainWindow open.
        """
        if not os.path.exists(project_dirpath):
            raise ValueError('Project does not exist: ' + project_dirpath)
        
        if readonly is not None:
            self.open_as_readonly.Value = readonly
        
        if using_crystalopen:
            itempath_to_open = os.path.join(project_dirpath, Project._OPENER_DEFAULT_FILENAME)
        else:
            itempath_to_open = project_dirpath
        
        with file_dialog_returning(itempath_to_open):
            click_button(self.open_button)
            
            if wait_func is not None:
                await wait_func()
            mw = await MainWindow.wait_for(timeout=self._TIMEOUT_FOR_OPEN_MAIN_WINDOW)
        
        project = Project._last_opened_project
        assert project is not None
        
        exc_info_while_close = None
        try:
            yield (mw, project)
        except:
            exc_info_while_close = sys.exc_info()
            raise
        finally:
            if autoclose:
                await mw.close(exc_info_while_close)
    
    async def start_opening(self, project_dirpath: str, *, next_window_name: str) -> None:
        with file_dialog_returning(project_dirpath):
            click_button(self.open_button)
            
            await wait_for(
                or_condition(
                    window_condition('cr-opening-project'),
                    window_condition('cr-main-window'),
                    window_condition(next_window_name)),
                stacklevel_extra=1)


class MainWindow:
    _connect_timeout: float | None
    
    main_window: wx.Frame
    entity_tree: EntityTree
    new_root_url_button: wx.Button
    new_group_button: wx.Button
    edit_button: wx.Button
    forget_button: wx.Button
    download_button: wx.Button
    update_members_button: wx.Button
    view_button: wx.Button
    task_tree: wx.TreeCtrl
    preferences_button: wx.Button
    read_write_icon: wx.StaticText
    
    @staticmethod
    async def wait_for(*, timeout: float | None=None) -> MainWindow:
        self = MainWindow(ready=True)
        self._connect_timeout = timeout
        await self._connect()
        return self
    
    async def _connect(self) -> None:
        self.main_window = await wait_for(
            window_condition('cr-main-window'),
            timeout=self._connect_timeout,
            stacklevel_extra=1)
        assert isinstance(self.main_window, wx.Frame)
        entity_tree_window = self.main_window.FindWindow(name='cr-entity-tree')
        assert isinstance(entity_tree_window, wx.TreeCtrl)
        self.entity_tree = EntityTree(entity_tree_window)
        self.new_root_url_button = self.main_window.FindWindow(name='cr-add-url-button')
        assert isinstance(self.new_root_url_button, wx.Button)
        self.new_group_button = self.main_window.FindWindow(name='cr-add-group-button')
        assert isinstance(self.new_group_button, wx.Button)
        self.edit_button = self.main_window.FindWindow(name='cr-edit-button')
        assert isinstance(self.edit_button, wx.Button)
        self.forget_button = self.main_window.FindWindow(name='cr-forget-button')
        assert isinstance(self.forget_button, wx.Button)
        self.download_button = self.main_window.FindWindow(name='cr-download-button')
        assert isinstance(self.download_button, wx.Button)
        self.update_members_button = self.main_window.FindWindow(name='cr-update-members-button')
        assert isinstance(self.update_members_button, wx.Button)
        self.view_button = self.main_window.FindWindow(name='cr-view-button')
        assert isinstance(self.view_button, wx.Button)
        self.task_tree = self.main_window.FindWindow(name='cr-task-tree')
        assert isinstance(self.task_tree, wx.TreeCtrl)
        self.preferences_button = self.main_window.FindWindow(name='cr-preferences-button')
        assert isinstance(self.preferences_button, wx.Button)
        self.read_write_icon = self.main_window.FindWindow(name='cr-read-write-icon')
        assert isinstance(self.read_write_icon, wx.StaticText)
    
    def __init__(self, *, ready: bool=False) -> None:
        assert ready, 'Did you mean to use MainWindow.wait_for()?'
    
    # === Menubar ===
    
    @property
    def entity_menu(self) -> wx.Menu:
        mb = self.main_window.MenuBar
        entity_menu_index = mb.FindMenu('Entity')
        assert entity_menu_index != wx.NOT_FOUND
        entity_menu = mb.GetMenu(entity_menu_index)
        return entity_menu
    
    # === Properties ===
    
    @property
    def readonly(self) -> bool:
        label = self.read_write_icon.Label  # cache
        if label in ['🔒', 'Read only']:
            return True
        elif label in ['✏️', 'Writable']:
            return False
        else:
            raise AssertionError()
    
    # === Operations ===
    
    async def click_download_button(self, *, immediate_finish_ok: bool=False) -> None:
        """
        Clicks the "Download" button and waits for it to finish starting
        a download task.
        """
        task_root_ti = TreeItem.GetRootItem(self.task_tree)
        assert task_root_ti is not None
        
        old_task_count = len(task_root_ti.Children)
        
        click_button(self.download_button)
        def task_count_changed_condition() -> int | None:
            assert task_root_ti is not None
            new_task_count = len(task_root_ti.Children)
            if new_task_count != old_task_count:
                return new_task_count
            else:
                return None
        try:
            await wait_for(task_count_changed_condition, stacklevel_extra=1)
        except WaitTimedOut:
            if immediate_finish_ok:
                # TOOD: We ended up waiting until the entire timeout expired.
                #       Find a more efficient way to detect that the download
                #       started and immediately finished.
                return
            else:
                raise
    
    # === Close ===
    
    CLOSE_TIMEOUT = 4.0
    
    async def close(self, exc_info=None) -> None:
        if is_synced_with_scheduler_thread():
            # Immediately close project
            pass
        else:
            # Try wait for any lingering tasks to complete.
            # 
            # Does workaround: https://github.com/davidfstr/Crystal-Web-Archiver/issues/74
            try:
                await wait_for(
                    tree_has_no_children_condition(self.task_tree),
                    timeout=self.CLOSE_TIMEOUT,  # wait only briefly
                    stacklevel_extra=1)
            except WaitTimedOut:
                first_task_title = first_task_title_progression(self.task_tree)()
                print(f'*** MainWindow: Closing while tasks are still running. May deadlock! Current task: {first_task_title}')
                
                # Print traceback or current exception that is being handled immediately
                # because impending deadlock may prevent traceback from being printed
                # in the usual manner
                if exc_info is None:
                    traceback.print_stack()
                else:
                    traceback.print_exception(*exc_info)
                
                # (continue)
        
        self.main_window.Close()
        await self.wait_for_close()
    
    async def close_with_menuitem(self) -> None:
        close_menuitem = self.main_window.MenuBar.FindItemById(wx.ID_CLOSE)
        wx.PostEvent(close_menuitem.Menu, wx.CommandEvent(wx.EVT_MENU.typeId, close_menuitem.Id))
        await self.wait_for_close()
    
    async def wait_for_close(self) -> None:
        await wait_for(lambda: self.main_window.IsBeingDeleted)
        await wait_for(lambda: not self.main_window.IsShown)
        await wait_for(
            not_condition(window_condition('cr-main-window')),
            timeout=4.0,  # 2.0s isn't long enough for macOS test runners on GitHub Actions
            stacklevel_extra=1,
        )
        await OpenOrCreateDialog.wait_for(
            timeout=4.0  # 2.0s isn't long enough for macOS test runners on GitHub Actions
        )
    
    @asynccontextmanager
    async def temporarily_closed(self, project_dirpath: str) -> AsyncIterator[None]:
        await self.close()
        try:
            yield
        finally:
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath, autoclose=False):
                pass
            await self._connect()  # reconnect self
    
    async def open_preferences_with_menuitem(self) -> PreferencesDialog:
        prefs_menuitem = self.main_window.MenuBar.FindItemById(wx.ID_PREFERENCES)
        wx.PostEvent(self.main_window, wx.CommandEvent(wx.EVT_MENU.typeId, prefs_menuitem.Id))
        return await PreferencesDialog.wait_for()
    
    async def quit_with_menuitem(self) -> None:
        quit_menuitem = self.main_window.MenuBar.FindItemById(wx.ID_EXIT)
        wx.PostEvent(self.main_window, wx.CommandEvent(wx.EVT_MENU.typeId, quit_menuitem.Id))
        
        await self.wait_for_close()


class NewRootUrlDialog:
    _dialog: wx.Dialog
    _controller: RealNewRootUrlDialog
    url_field: wx.TextCtrl
    url_cleaner_spinner: wx.ActivityIndicator
    name_field: wx.TextCtrl
    ok_button: wx.Button
    cancel_button: wx.Button
    
    download_immediately_checkbox: wx.CheckBox  # or None
    create_group_checkbox: wx.CheckBox  # or None
    
    options_button: wx.Button
    set_as_default_domain_checkbox: wx.CheckBox
    set_as_default_directory_checkbox: wx.CheckBox
    
    @staticmethod
    async def wait_for() -> NewRootUrlDialog:
        self = NewRootUrlDialog(ready=True)
        self._dialog = await wait_for(window_condition('cr-new-root-url-dialog'), stacklevel_extra=1)
        assert isinstance(self._dialog, wx.Dialog)
        assert RealNewRootUrlDialog._last_opened is not None
        self._controller = RealNewRootUrlDialog._last_opened
        self.url_field = self._dialog.FindWindow(name='cr-new-root-url-dialog__url-field')
        assert isinstance(self.url_field, wx.TextCtrl)
        self.url_cleaner_spinner = self._dialog.FindWindow(name='cr-new-root-url-dialog__url-cleaner-spinner')
        assert isinstance(self.url_cleaner_spinner, wx.ActivityIndicator)
        self.name_field = self._dialog.FindWindow(name='cr-new-root-url-dialog__name-field')
        assert isinstance(self.name_field, wx.TextCtrl)
        self.ok_button = self._dialog.FindWindow(id=wx.ID_NEW)
        if self.ok_button is None:
            self.ok_button = self._dialog.FindWindow(id=wx.ID_SAVE)
        assert isinstance(self.ok_button, wx.Button)
        self.cancel_button = self._dialog.FindWindow(id=wx.ID_CANCEL)
        assert isinstance(self.cancel_button, wx.Button)
        
        self.download_immediately_checkbox = self._dialog.FindWindow(name='cr-new-root-url-dialog__download-immediately-checkbox')
        assert (
            self.download_immediately_checkbox is None or
            isinstance(self.download_immediately_checkbox, wx.CheckBox)
        )
        self.create_group_checkbox = self._dialog.FindWindow(name='cr-new-root-url-dialog__create-group-checkbox')
        assert (
            self.create_group_checkbox is None or
            isinstance(self.create_group_checkbox, wx.CheckBox)
        )
        
        self.options_button = self._dialog.FindWindow(id=wx.ID_MORE)
        assert isinstance(self.options_button, wx.Button)
        self.set_as_default_domain_checkbox = self._dialog.FindWindow(name='cr-new-root-url-dialog__set-as-default-domain-checkbox')
        assert isinstance(self.set_as_default_domain_checkbox, wx.CheckBox)
        self.set_as_default_directory_checkbox = self._dialog.FindWindow(name='cr-new-root-url-dialog__set-as-default-directory-checkbox')
        assert isinstance(self.set_as_default_directory_checkbox, wx.CheckBox)
        
        return self
    
    def __init__(self, *, ready: bool=False) -> None:
        assert ready, 'Did you mean to use NewRootUrlDialog.wait_for()?'
    
    @property
    def shown(self) -> bool:
        if self._controller._is_destroying_or_destroyed:
            return False
        return self._dialog.IsShown()
    
    @property
    def url_field_focused(self) -> bool:
        return self._controller._url_field_focused
    
    @property
    def new_options_shown(self) -> bool:
        return (
            self.download_immediately_checkbox is not None and
            self.create_group_checkbox is not None
        )
    
    async def ok(self) -> None:
        click_button(self.ok_button)
        await wait_for(not_condition(window_condition('cr-new-root-url-dialog')), stacklevel_extra=1)
    
    async def cancel(self) -> None:
        click_button(self.cancel_button)
        await wait_for(not_condition(window_condition('cr-new-root-url-dialog')), stacklevel_extra=1)
    
    # === Utility ===
    
    def do_not_download_immediately(self) -> None:
        """
        Configures the URL being created so that it is NOT immediately downloaded
        after creation.
        
        Several tests that create a URL are not interested in the default
        "download immediately" behavior and are simpler to write when there
        is no need to worry about or clean up after a URL is downloaded as
        a side effect of creating it.
        """
        if self.download_immediately_checkbox is None:
            return
        if self.download_immediately_checkbox.Value:
            self.download_immediately_checkbox.Value = False
    
    def do_not_set_default_url_prefix(self) -> None:
        """
        Configures the URL being created to NOT also set it as the default domain.
        
        Several test infrastructure methods (like TreeItem.find_child) are easier
        to use when there is no Default URL Prefix set, so some tests intentionally
        avoid setting a Default URL Prefix.
        """
        click_button(self.options_button)
        if self.set_as_default_domain_checkbox.Value:
            self.set_as_default_domain_checkbox.Value = False
        
        assert False == self.set_as_default_domain_checkbox.Value
        assert False == self.set_as_default_directory_checkbox.Value


class NewGroupDialog:
    _NONE_SOURCE_TITLE = 'none'
    _SOURCE_TITLE_RE = re.compile(r'^(?:[^a-zA-Z0-9]+ )?(.*?)(?: - (.*?))? *$')
    
    name_field: wx.TextCtrl
    pattern_field: wx.TextCtrl
    source_field: wx.Choice
    preview_members_pane: wx.CollapsiblePane | None
    preview_members_list: wx.ListBox
    cancel_button: wx.Button
    ok_button: wx.Button
    
    download_immediately_checkbox: wx.CheckBox  # or None
    
    @staticmethod
    async def wait_for() -> NewGroupDialog:
        self = NewGroupDialog(ready=True)
        add_group_dialog = await wait_for(
            NewGroupDialog.window_condition(),
            timeout=5.0,  # 4.2s observed for macOS test runners on GitHub Actions
            stacklevel_extra=1,
        )  # type: wx.Window
        self.name_field = add_group_dialog.FindWindow(name='cr-new-group-dialog__name-field')
        assert isinstance(self.name_field, wx.TextCtrl)
        self.pattern_field = add_group_dialog.FindWindow(name='cr-new-group-dialog__pattern-field')
        assert isinstance(self.pattern_field, wx.TextCtrl)
        self.source_field = add_group_dialog.FindWindow(name='cr-new-group-dialog__source-field')
        assert isinstance(self.source_field, wx.Choice)
        self.preview_members_pane = add_group_dialog.FindWindow(name='cr-new-group-dialog__preview-members')
        assert (
            self.preview_members_pane is None or
            isinstance(self.preview_members_pane, wx.CollapsiblePane)
        )
        self.preview_members_list = add_group_dialog.FindWindow(name='cr-new-group-dialog__preview-members__list')
        assert isinstance(self.preview_members_list, wx.ListBox)
        self.cancel_button = add_group_dialog.FindWindow(id=wx.ID_CANCEL)
        assert isinstance(self.cancel_button, wx.Button)
        self.ok_button = add_group_dialog.FindWindow(id=wx.ID_NEW)
        if self.ok_button is None:
            self.ok_button = add_group_dialog.FindWindow(id=wx.ID_SAVE)
        assert isinstance(self.ok_button, wx.Button)
        
        self.download_immediately_checkbox = add_group_dialog.FindWindow(name='cr-new-group-dialog__download-immediately-checkbox')
        assert (
            self.download_immediately_checkbox is None or
            isinstance(self.download_immediately_checkbox, wx.CheckBox)
        )
        
        return self
    
    @staticmethod
    def window_condition() -> Callable[[], wx.Window | None]:
        return window_condition('cr-new-group-dialog')
    
    def __init__(self, *, ready: bool=False) -> None:
        assert ready, 'Did you mean to use NewGroupDialog.wait_for()?'
    
    # TODO: Rename -> source_name
    def _get_source(self) -> str | None:
        selection_ci = self.source_field.GetSelection()
        if selection_ci == wx.NOT_FOUND:
            return None
        source_title = self.source_field.GetString(selection_ci)
        if source_title == self._NONE_SOURCE_TITLE:
            return None
        m = self._SOURCE_TITLE_RE.fullmatch(source_title)
        assert m is not None
        source_name = m.group(2)
        if source_name is not None:
            return source_name
        else:
            # If the referenced source has no name, allow referring to it by its display URL
            cur_source_display_url = m.group(1)
            assert cur_source_display_url is not None
            return cur_source_display_url
            
    def _set_source(self, source_name: str | None) -> None:
        if source_name is None:
            selection_ci = 0
        else:
            for (selection_ci, source_title) in enumerate(self.source_field.GetStrings()):
                if selection_ci == 0:
                    continue
                m = self._SOURCE_TITLE_RE.fullmatch(source_title)
                assert m is not None
                cur_source_name = m.group(2)
                if cur_source_name is not None:
                    if cur_source_name == source_name:
                        break
                else:
                    # If the referenced source has no name, allow referring to it by its display URL
                    cur_source_display_url = m.group(1)
                    assert cur_source_display_url is not None
                    if cur_source_display_url == source_name:
                        break
            else:
                raise ValueError(f'Source not found: {source_name}')
        self.source_field.SetSelection(selection_ci)
    source = property(_get_source, _set_source)
    
    @property
    def new_options_shown(self) -> bool:
        return (
            self.download_immediately_checkbox is not None
        )
    
    async def ok(self) -> None:
        click_button(self.ok_button)
        await wait_for(not_condition(window_condition('cr-new-group-dialog')), stacklevel_extra=1)
    
    async def cancel(self) -> None:
        click_button(self.cancel_button)
        await wait_for(not_condition(window_condition('cr-new-group-dialog')), stacklevel_extra=1)


class PreferencesDialog:
    html_parser_field: wx.Choice
    stale_before_checkbox: wx.CheckBox
    stale_before_date_picker: wx.adv.DatePickerCtrl
    cookie_field: wx.ComboBox
    ok_button: wx.Button
    
    @staticmethod
    async def wait_for() -> PreferencesDialog:
        import wx.adv
        
        self = PreferencesDialog(ready=True)
        preferences_dialog = await wait_for(
            window_condition('cr-preferences-dialog'),
            timeout=4.0,  # 2.0s isn't long enough for macOS test runners on GitHub Actions
            stacklevel_extra=1,
        )  # type: wx.Window
        self.html_parser_field = preferences_dialog.FindWindow(name=
            'cr-preferences-dialog__html-parser-field')
        assert isinstance(self.html_parser_field, wx.Choice)
        self.stale_before_checkbox = preferences_dialog.FindWindow(name=
            'cr-preferences-dialog__stale-before-checkbox')
        assert isinstance(self.stale_before_checkbox, wx.CheckBox)
        self.stale_before_date_picker = preferences_dialog.FindWindow(name=
            'cr-preferences-dialog__stale-before-date-picker')
        assert isinstance(self.stale_before_date_picker, wx.adv.DatePickerCtrl)
        self.cookie_field = preferences_dialog.FindWindow(name=
            'cr-preferences-dialog__cookie-field')
        assert isinstance(self.cookie_field, wx.ComboBox)
        self.ok_button = preferences_dialog.FindWindow(id=wx.ID_OK)
        assert isinstance(self.ok_button, wx.Button)
        return self
    
    def __init__(self, *, ready: bool=False) -> None:
        assert ready, 'Did you mean to use PreferencesDialog.wait_for()?'
    
    async def ok(self) -> None:
        click_button(self.ok_button)
        await wait_for(not_condition(window_condition('cr-preferences-dialog')), stacklevel_extra=1)


# ------------------------------------------------------------------------------
# Panel Abstractions

class EntityTree:
    def __init__(self, window: wx.TreeCtrl) -> None:
        self.window = window
    
    async def get_tree_item_icon_tooltip(self, tree_item: TreeItem) -> str | None:
        if tree_item.tree != self.window:
            raise ValueError()
        return tree_item.Tooltip('icon')
    
    @staticmethod
    async def assert_tree_item_icon_tooltip_contains(ti: TreeItem, value: str) -> None:
        tooltip = await (EntityTree(ti.tree).get_tree_item_icon_tooltip(ti))
        assert tooltip is not None and value in tooltip, \
            f'Expected tooltip to contain {value!r}, but it was: {tooltip!r}'
    
    async def set_default_domain_to_entity_at_tree_item(self, tree_item: TreeItem) -> None:
        """
        Raises:
        * MenuitemMissingError
        * MenuitemDisabledError
        """
        await self._choose_action_for_entity_at_tree_item(tree_item, 'Set As Default Domain')
    
    async def set_default_directory_to_entity_at_tree_item(self, tree_item: TreeItem) -> None:
        """
        Raises:
        * MenuitemMissingError
        * MenuitemDisabledError
        """
        await self._choose_action_for_entity_at_tree_item(tree_item, 'Set As Default Directory')
    
    async def clear_default_domain_from_entity_at_tree_item(self, tree_item: TreeItem) -> None:
        """
        Raises:
        * MenuitemMissingError
        * MenuitemDisabledError
        """
        await self._choose_action_for_entity_at_tree_item(tree_item, 'Clear Default Domain')
    
    async def clear_default_directory_from_entity_at_tree_item(self, tree_item: TreeItem) -> None:
        """
        Raises:
        * MenuitemMissingError
        * MenuitemDisabledError
        """
        await self._choose_action_for_entity_at_tree_item(tree_item, 'Clear Default Directory')
    
    async def _choose_action_for_entity_at_tree_item(self,
            tree_item: TreeItem,
            action_prefix: str) -> None:
        """
        Raises:
        * MenuitemMissingError
        * MenuitemDisabledError
        """
        def show_popup(menu: wx.Menu) -> None:
            try:
                (set_prefix_menuitem,) = (
                    mi for mi in menu.MenuItems
                    if mi.ItemLabelText.startswith(action_prefix)
                )
            except ValueError:  # not enough values to unpack
                raise MenuitemMissingError()
            if not set_prefix_menuitem.Enabled:
                raise MenuitemDisabledError()
            select_menuitem_now(menu, set_prefix_menuitem.Id)
        await tree_item.right_click_showing_popup_menu(show_popup)


class MenuitemMissingError(ValueError):
    pass


class MenuitemDisabledError(ValueError):
    pass


# ------------------------------------------------------------------------------



from crystal import progress
import crystal.main
from crystal.model import Project, Resource
from crystal.progress import CancelOpenProject
from crystal.tests.util.controls import TreeItem
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.server import extracted_project
from crystal.tests.util.skip import skipTest
from crystal.tests.util.ssd import database_on_ssd
from crystal.tests.util.subtests import awith_subtests, SubtestsContext
from crystal.tests.util.wait import (
    first_child_of_tree_item_is_not_loading_condition, wait_for,
)
from crystal.tests.util.windows import OpenOrCreateDialog
from crystal.util.xos import is_linux, is_mac_os, is_windows
import os.path
import tempfile
from unittest import skip
from unittest.mock import call, patch
import wx

# === Test: Create Project ===

async def test_when_create_project_then_shows_dialog_saying_opening_project_but_no_other_messages() -> None:
    ocd = await OpenOrCreateDialog.wait_for()
    
    progress_listener = progress._active_progress_listener
    assert progress_listener is not None
    
    with patch('crystal.progress._DELAY_UNTIL_PROGRESS_DIALOG_SHOWS', 0), \
            patch('crystal.progress.wx.ProgressDialog', autospec=True) as MockProgressDialog:
        pd = MockProgressDialog.return_value
        pd.Pulse.return_value = (True, False)
        pd.Update.return_value = (True, False)
        
        async with ocd.create() as (mw, _):
            pass
        
        assert (
            [call('Opening project...')],
            []
        ) == (
            pd.Pulse.call_args_list,
            pd.Update.call_args_list
        )


async def test_can_create_project_with_url_unsafe_characters() -> None:
    with tempfile.TemporaryDirectory(prefix='original#2', suffix='.crystalproj') as project_dirpath:
        assert '#' in project_dirpath
        
        ocd = await OpenOrCreateDialog.wait_for()
        async with ocd.create(project_dirpath) as (mw, _):
            assert os.path.exists(os.path.join(project_dirpath, Project._DB_FILENAME))


# === Test: Start Open Project ===

@skip('not yet automated: hard to automate')
async def test_given_macos_when_double_click_crystalproj_package_in_finder_then_opens_project() -> None:
    pass


async def test_given_macos_when_open_crystalproj_package_in_open_dialog_then_opens_project() -> None:
    if not is_mac_os():
        skipTest('only supported on macOS')
    
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        ocd = await OpenOrCreateDialog.wait_for()
        async with ocd.open(project_dirpath, using_crystalopen=False) as (mw, project):
            pass


@skip('not yet automated: hard to automate')
async def test_given_windows_when_double_click_crystalproj_directory_in_explorer_then_opens_project() -> None:
    pass


@skip('not yet automated: hard to automate')
async def test_given_windows_when_open_crystalproj_directory_and_double_click_crystalopen_file_in_explorer_then_opens_project() -> None:
    pass


async def test_given_windows_when_open_crystalproj_directory_and_double_click_crystalopen_file_in_open_dialog_then_opens_project() -> None:
    if not is_windows():
        skipTest('only supported on Windows')
    
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        with Project(project_dirpath):
            pass
        assert os.path.exists(os.path.join(project_dirpath, Project._OPENER_DEFAULT_FILENAME))
        
        ocd = await OpenOrCreateDialog.wait_for()
        async with ocd.open(project_dirpath, using_crystalopen=True) as (mw, project):
            pass


@skip('not yet automated: hard to automate')
async def test_given_linux_when_open_crystalproj_directory_and_double_click_crystalopen_file_in_file_explorer_then_opens_project() -> None:
    pass


async def test_given_linux_when_open_crystalproj_directory_in_open_dialog_then_opens_project() -> None:
    if not is_linux():
        skipTest('only supported on Linux')
    
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        assert not os.path.exists(os.path.join(project_dirpath, Project._OPENER_DEFAULT_FILENAME))
        
        # Simulate effect of:
        # 1. Press "Open" button to open the "Choose a project" dialog
        # 2. Select a .crystalproj directory. "Open Directory" button undims.
        # 3. Press "Open Directory" button
        ocd = await OpenOrCreateDialog.wait_for()
        async with ocd.open(project_dirpath, using_crystalopen=False) as (mw, project):
            pass


async def test_given_linux_when_open_crystalproj_directory_and_double_click_crystalopen_file_in_open_dialog_then_opens_project() -> None:
    if not is_linux():
        skipTest('only supported on Linux')
    
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        with Project(project_dirpath):
            pass
        assert os.path.exists(os.path.join(project_dirpath, Project._OPENER_DEFAULT_FILENAME))
        
        ocd = await OpenOrCreateDialog.wait_for()
        async with ocd.open(project_dirpath, using_crystalopen=True) as (mw, project):
            pass


# === Test: While Opening Project ===

@awith_subtests
async def test_given_project_opening_when_click_cancel_then_returns_to_prompt_dialog(
        subtests: SubtestsContext) -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        for method_name in [
                # Case 1: Cancel while creating Project object
                'loading_root_resources',
                # Case 2: Cancel while creating MainWindow object
                'creating_entity_tree_nodes',
                ]:
            with subtests.test(method_name=method_name):
                ocd = await OpenOrCreateDialog.wait_for()
                
                progress_listener = progress._active_progress_listener
                assert progress_listener is not None
                
                with patch.object(progress_listener, method_name, side_effect=CancelOpenProject):
                    await ocd.start_opening(project_dirpath, next_window_name='cr-open-or-create-project')
                    
                    # HACK: Wait minimum duration to allow open to finish
                    await bg_sleep(0.5)
                    
                    ocd = await OpenOrCreateDialog.wait_for()


async def test_given_project_is_corrupt_when_open_project_then_displays_error_dialog() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Introduce corruption to the project
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            group = list(project.resource_groups)[0]  # arbitrary
            
            c = project._db.cursor()
            c.execute(
                'update resource_group set source_type = ? where id = ?',
                ('bogus', group._id))
            project._db.commit()
        
        # Try to open that project
        if True:
            # Prepare to: OK
            did_respond_to_invalid_project_dialog = False
            def click_ok_in_invalid_project_dialog(dialog: wx.Dialog) -> int:
                assert 'cr-invalid-project' == dialog.Name
                assert 'may be corrupted' in dialog.Message
                
                nonlocal did_respond_to_invalid_project_dialog
                did_respond_to_invalid_project_dialog = True
                
                return wx.ID_OK
            
            load_project = crystal.main._load_project  # capture
            def patched_load_project(*args, **kwargs) -> Project:
                return load_project(
                    *args,
                    _show_modal_func=click_ok_in_invalid_project_dialog,  # type: ignore[misc]
                    **kwargs)
            
            with patch(f'crystal.main._load_project', patched_load_project):
                ocd = await OpenOrCreateDialog.wait_for()
                await ocd.start_opening(project_dirpath, next_window_name='cr-open-or-create-project')
                
                # HACK: Wait minimum duration to allow open to finish
                await bg_sleep(0.5)
                
                # Wait for cancel and return to initial dialog
                ocd = await OpenOrCreateDialog.wait_for()
                
                assert did_respond_to_invalid_project_dialog


# === Test: After Opening Project ===

async def test_given_project_was_just_opened_then_first_entity_selected() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        home_url = 'https://xkcd.com/'
        
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            home_ti = root_ti.find_child(home_url)
            assert home_ti.IsSelected()
            
            # Ensure can view the downloaded site with 1 click
            assert mw.view_button.Enabled


@awith_subtests
async def test_given_project_was_just_opened_then_no_resources_loaded_except_root_resources(subtests: SubtestsContext) -> None:
    for is_ssd in [False, True]:
        with subtests.test(is_ssd=is_ssd), database_on_ssd(is_ssd):
            with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
                async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                    # Ensure project opens with only root resources loaded
                    assert len(list(project.root_resources)) >= 1
                    resources_that_are_loaded = _sort_resources(
                        _loaded_resources(project))  # capture
                    resources_that_are_root = _sort_resources(
                        [rr.resource for rr in project.root_resources])
                    assert resources_that_are_root == resources_that_are_loaded


async def test_given_on_ssd_when_resource_group_node_expanded_then_only_new_resources_loaded_are_group_members() -> None:
    with database_on_ssd(True):
        with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                # Ensure when group expanded, only that group's members become loaded
                if True:
                    comic_group = project.get_resource_group('Comics')
                    assert comic_group is not None
                    
                    root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                    
                    (comic_group_ti,) = (
                        child for child in root_ti.Children
                        if child.Text.endswith(f'- {comic_group.name}')
                    )
                    
                    comic_group_ti.Expand()
                    await wait_for(first_child_of_tree_item_is_not_loading_condition(comic_group_ti))
                    
                    assert len(list(comic_group.members)) >= 1
                    resources_that_are_loaded = _sort_resources(
                        _loaded_resources(project))  # capture
                    resources_that_are_root_or_group_members = _sort_resources(
                        [rr.resource for rr in project.root_resources] + list(comic_group.members))
                    assert resources_that_are_root_or_group_members == resources_that_are_loaded, (
                        f'{resources_that_are_root_or_group_members=} but {resources_that_are_loaded=}'
                    )


async def test_given_not_on_ssd_when_resource_group_node_expanded_then_all_project_resources_are_loaded() -> None:
    with database_on_ssd(False):
        with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                # Ensure when group expanded, all project resources become loaded
                if True:
                    comic_group = project.get_resource_group('Comics')
                    assert comic_group is not None
                    
                    root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                    
                    (comic_group_ti,) = (
                        child for child in root_ti.Children
                        if child.Text.endswith(f'- {comic_group.name}')
                    )
                    
                    comic_group_ti.Expand()
                    await wait_for(first_child_of_tree_item_is_not_loading_condition(comic_group_ti))
                    
                    assert len(list(comic_group.members)) >= 1
                    resources_that_are_loaded = _sort_resources(
                        _loaded_resources(project))  # capture
                    resources_in_project = _sort_resources(
                        list(project.resources))
                    
                    assert resources_in_project == resources_that_are_loaded, (
                        f'{resources_in_project=} but {resources_that_are_loaded=}'
                    )


def _loaded_resources(project: Project) -> list[Resource]:
    # HACK: Use private API
    resource_for_url = project._resource_for_url
    resource_for_id = project._resource_for_id
    assert len(resource_for_url) == len(resource_for_id)
    return list(resource_for_url.values())


def _sort_resources(resources: list[Resource]) -> list[Resource]:
    return sorted(resources, key=lambda r: r._id)



"""
Parses XML documents.
"""

from bs4 import BeautifulSoup
from crystal.doc.generic import Document, Link
from crystal.doc.html.soup import HtmlDocument, HtmlLink
from crystal.util.fastsoup import BeautifulFastSoup
from typing import BinaryIO, List


def parse_xml_and_links(
        xml_bytes: BinaryIO,
        declared_charset: str | None=None
        ) -> tuple[Document, list[Link]] | None:
    """
    Parses an XML document, returning a FastSoup object that can be
    examined through a BeautifulSoup-compatible API.
    
    Returns None if there was a parsing error.
    """
    try:
        xml = BeautifulFastSoup(BeautifulSoup(
            xml_bytes,
            from_encoding=declared_charset,
            features='xml',
        ))
    except Exception:
        return None
    
    links = []
    
    # <link>*</link> (from RSS feed)
    # <link href=*> (from Atom feed)
    type_title = 'Link'
    title = None
    embedded = False
    for tag in xml.find_all('link'):
        if xml.tag_string(tag) not in [None, '']:
            links.append(HtmlLink.create_from_tag(tag, xml, 'string', type_title, title, embedded))
        if 'href' in  xml.tag_attrs(tag):  # usually also has: rel="alternate"
            links.append(HtmlLink.create_from_tag(tag, xml, 'href', type_title, title, embedded))
    
    links_ = links  # type: List[Link]  # type: ignore[assignment]  # allow List[HtmlLink] to be converted
    return (HtmlDocument(xml, is_html=False), links_)




import atexit
from collections.abc import Callable, Iterator
from contextlib import AbstractContextManager, contextmanager, nullcontext
import cProfile
from functools import wraps
import inspect
import sys
import threading
import time
from typing import Optional

_excluded_delta_time_stack = threading.local()

@contextmanager
def warn_if_slow(
        title: str,
        max_duration: float,
        message: Callable[[], str] | str,
        *, enabled: bool=True
        ) -> Iterator[None]:
    """
    Context that profiles the runtime of its enclosed code and
    upon exit prints a warning if the runtime exceeds the specified `max_duration`.
    
    If a warn_if_slow() context is nested within another warn_if_slow() context,
    the runtime of the inner context will be excluded from the runtime of the
    outer context.
    
    If a ignore_runtime_from_enclosing_warn_if_slow() context is nested within
    a warn_if_slow() context, the runtime of the inner context will be excluded
    from the runtime of the outer context.
    """
    if not enabled:
        yield
        return
    if not (max_duration > 0):
        raise ValueError()
    
    if not hasattr(_excluded_delta_time_stack, 'value'):
        _excluded_delta_time_stack.value = []
    _excluded_delta_time_stack.value.append(0)
    
    start_time = time.time()  # capture
    try:
        yield
    finally:
        end_time = time.time()  # capture
        
        excluded_delta_time = _excluded_delta_time_stack.value.pop()
        
        delta_time = end_time - start_time  # cache
        if (delta_time - excluded_delta_time) > max_duration:
            message_str = message() if callable(message) else message
            assert isinstance(message_str, str)
            excluded_part = (
                ' ({:.02f}s excluded)'.format(excluded_delta_time)
                if excluded_delta_time > 0
                else ''
            )
            print("*** {} took {:.02f}s{} to execute: {}".format(
                title,
                delta_time,
                excluded_part,
                message_str,
            ), file=sys.stderr)
            
            # Exclude delta_time from any enclosing calls of warn_if_slow()
            for (i, _) in enumerate(_excluded_delta_time_stack.value):
                _excluded_delta_time_stack.value[i] += delta_time


@contextmanager
def ignore_runtime_from_enclosing_warn_if_slow() -> Iterator[None]:
    """
    Context that excluded its runtime from that of any enclosing
    warn_if_slow() context.
    """
    if not hasattr(_excluded_delta_time_stack, 'value'):
        _excluded_delta_time_stack.value = []
    
    start_time = time.time()  # capture
    try:
        yield
    finally:
        end_time = time.time()  # capture
        delta_time = end_time - start_time  # cache
        
        # Exclude delta_time from any enclosing calls of warn_if_slow()
        for (i, _) in enumerate(_excluded_delta_time_stack.value):
            _excluded_delta_time_stack.value[i] += delta_time


def create_profiled_callable(title: str, max_duration: float, callable: Callable, *args) -> Callable:
    """
    Decorates the specified callable such that it prints
    a warning to the console if its runtime is long.
    """
    @wraps(callable)
    def profiled_callable() -> None:
        def message_func() -> str:
            # TODO: Remove support for nested callable objects that are
            #       tracked using a "callable" attribute. I believe this
            #       is dead code.
            root_callable = callable
            while hasattr(root_callable, 'callable'):
                root_callable = root_callable.callable  # type: ignore[attr-defined]
            
            try:
                file = inspect.getsourcefile(root_callable)
            except Exception:
                file = '?'
            try:
                start_line_number = str(inspect.getsourcelines(root_callable)[-1])
            except Exception:
                start_line_number = '?'
            return '{} @ [{}:{}]'.format(
                root_callable,
                file,
                start_line_number
            )
            
        with warn_if_slow(title, max_duration, message_func):
            return callable(*args)  # cr-traceback: ignore
    return profiled_callable


def create_profiling_context(
        stats_filepath: str,
        *, enabled: bool=True,
        ) -> 'AbstractContextManager[Optional[cProfile.Profile]]':
    """
    Creates a cProfile profiling context.
    Within the context all function calls are timed.
    The content can be entered and exited mutliple times.
    
    Just before the program exits, a .stats file is written to the specified
    filepath. This file can be analyzed/visualized with the PyPI "flameprof"
    module and the standard library "pstats" module.
    """
    if enabled:
        profiling_context = cProfile.Profile()  # type: AbstractContextManager[Optional[cProfile.Profile]]
        @atexit.register
        def dump_stats() -> None:  # type: ignore[misc]
            profiler = profiling_context
            assert isinstance(profiler, cProfile.Profile)
            profiler.dump_stats(stats_filepath)
    else:
        profiling_context = nullcontext(enter_result=None)
    return profiling_context



"""
A FinderInfo is a structure used to store certain macOS-specific file metadata.

References:
- http://dubeiko.com/development/FileSystems/HFSPLUS/tn1150.html#FinderInfo
- https://arstechnica.com/gadgets/2013/10/os-x-10-9/9/
"""
from crystal.util.xos import is_mac_os

_EMPTY = b'\x00' * 32


# Empirically-determined bit that is set when "Hide extension" is enabled
# for a file in the Finder
HIDE_FILE_EXTENSION = (
    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00'
    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
)


# === Get/Set FinderInfo ===

def get(filepath: str) -> bytes:
    assert is_mac_os()
    
    import xattr  # don't import at top-level because only available on macOS
    
    try:
        return xattr.getxattr(filepath, xattr.XATTR_FINDERINFO_NAME)
    except OSError as e:
        if e.errno == 93:  # 'Attribute not found'
            return _EMPTY
        else:
            raise


def set(filepath: str, finderinfo: bytes) -> None:
    assert is_mac_os()
    
    import xattr  # don't import at top-level because only available on macOS
    
    if finderinfo == _EMPTY:
        try:
            xattr.removexattr(filepath, xattr.XATTR_FINDERINFO_NAME)
        except OSError as e:
            if e.errno == 93:  # 'Attribute not found'
                pass
            else:
                raise
    else:
        xattr.setxattr(filepath, xattr.XATTR_FINDERINFO_NAME, finderinfo)


# === Hide File Extension ===

def set_hide_file_extension(itempath: str, hide: bool) -> None:
    old_fi = get(itempath)
    if hide:
        new_fi = or_(old_fi, HIDE_FILE_EXTENSION)
    else:
        new_fi = and_(old_fi, not_(HIDE_FILE_EXTENSION))
    set(itempath, new_fi)


def get_hide_file_extension(itempath: str) -> bool:
    fi = get(itempath)
    return and_(fi, HIDE_FILE_EXTENSION) != _EMPTY


# === Bitwise Operations ===

def and_(finderinfo1: bytes, finderinfo2: bytes) -> bytes:
    return bytes([a & b for (a, b) in zip(finderinfo1, finderinfo2)])

def or_(finderinfo1: bytes, finderinfo2: bytes) -> bytes:
    return bytes([a | b for (a, b) in zip(finderinfo1, finderinfo2)])

def not_(finderinfo: bytes) -> bytes:
    return bytes([a ^ 0b11111111 for a in finderinfo])



import sys
from typing import TYPE_CHECKING

__all__ = ['Ellipsis', 'EllipsisType']


# 1. Define EllipsisType, the type of Ellipsis
# 2. Define Ellipsis
if sys.version_info >= (3, 10):
    from types import EllipsisType
    Ellipsis = Ellipsis  # export
else:
    # https://github.com/python/typing/issues/684#issuecomment-548203158
    if TYPE_CHECKING:
        from enum import Enum
        class EllipsisType(Enum):
            Ellipsis = '...'
        Ellipsis = EllipsisType.Ellipsis
    else:
        EllipsisType = type(Ellipsis)
        Ellipsis = Ellipsis  # export


"""Tests for DownloadResourceTask and DownloadResourceGroupTask"""

from contextlib import redirect_stderr
from crystal.model import Project, Resource
import crystal.task
from crystal.tests.util.asserts import assertEqual
from crystal.tests.util.controls import click_button, TreeItem
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.server import MockHttpServer, served_project
from crystal.tests.util.tasks import wait_for_download_to_start_and_finish
from crystal.tests.util.wait import DEFAULT_WAIT_PERIOD, wait_for
from crystal.tests.util.windows import NewGroupDialog, OpenOrCreateDialog
import io
import os
import tempfile
from textwrap import dedent
from unittest import skip
from unittest.mock import patch

_FAVICON_PATH = '/favicon.ico'


# ------------------------------------------------------------------------------
# Resource.download() Tests

@skip('covered by: test_given_downloading_resource_when_start_download_resource_then_existing_download_task_returned')
async def test_given_not_downloading_resource_when_start_download_resource_then_download_task_created_and_returned() -> None:
    pass


async def test_given_downloading_resource_when_start_download_resource_then_existing_download_task_returned() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, home_url)
            
            rr_future = r.download()
            
            rr_future2 = r.download()
            assert rr_future2 is rr_future
            
            await wait_for_download_to_start_and_finish(mw.task_tree)


# ------------------------------------------------------------------------------
# DownloadResourceTask Tests

async def test_downloads_embedded_resources() -> None:
    server = MockHttpServer({
        '/': dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=dedent(
                """
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets/image.png" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8')
        )
    })
    with server:
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, server.get_url('/'))
            revision_future = r.download(wait_for_embedded=True)
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            assertEqual(['/', '/assets/image.png', _FAVICON_PATH], server.requested_paths)


async def test_does_not_download_embedded_resources_of_http_4xx_and_5xx_pages() -> None:
    server = MockHttpServer({
        '/': dict(
            status_code=404,
            headers=[('Content-Type', 'text/html')],
            content=dedent(
                """
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets/image.png" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8')
        ),
        '/assets/image.png': dict(
            status_code=404,
            headers=[('Content-Type', 'text/html')],
            content=dedent(
                """
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets/image.png" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8')
        )
    })
    with server:
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, server.get_url('/'))
            revision_future = r.download(wait_for_embedded=True)
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            assertEqual(['/'], server.requested_paths)


async def test_does_not_download_embedded_resources_of_recognized_binary_resource() -> None:
    server = MockHttpServer({
        '/': dict(
            status_code=200,
            headers=[('Content-Type', 'image/png')],
            content=dedent(
                """
                PNG <img src="/assets/image.png" />
                """
            ).lstrip('\n').encode('utf-8')
        )
    })
    with server:
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, server.get_url('/'))
            revision_future = r.download(wait_for_embedded=True)
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            assertEqual(['/'], server.requested_paths)


async def test_does_not_download_forever_given_embedded_resources_form_a_cycle() -> None:
    server = MockHttpServer({
        '/': dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=dedent(
                """
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets/image.png" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8')
        ),
        '/assets/image.png': dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=dedent(
                """
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets/image.png" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8')
        )
    })
    with server:
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, server.get_url('/'))
            revision_future = r.download(wait_for_embedded=True)
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            assertEqual(['/', '/assets/image.png', _FAVICON_PATH], server.requested_paths)


async def test_does_not_download_forever_given_embedded_resources_nest_infinitely() -> None:
    server = MockHttpServer({
        '/': dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=dedent(
                """
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets/image.png" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8')
        ),
        (lambda path: path.startswith('/assets/')): dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=(lambda path: dedent(
                f"""
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets{path}" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8'))
        )
    })
    with server:
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, server.get_url('/'))
            revision_future = r.download(wait_for_embedded=True)
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            assert 3 == crystal.task._MAX_EMBEDDED_RESOURCE_RECURSION_DEPTH
            assertEqual([
                '/',
                '/assets/image.png',  # 1
                '/assets/assets/image.png',  # 2
                '/assets/assets/assets/image.png',  # 3
                _FAVICON_PATH,
            ], server.requested_paths)


async def test_when_download_resource_given_revision_body_missing_then_redownloads_revision_body() -> None:
    server = MockHttpServer({
        '/': dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=dedent(
                """
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets/image.png" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8')
        ),
    })
    with server:
        with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
            async with (await OpenOrCreateDialog.wait_for()).create(project_dirpath) as (mw, project):
                r = Resource(project, server.get_url('/'))
                revision_future = r.download(wait_for_embedded=True)
                while not revision_future.done():
                    await bg_sleep(DEFAULT_WAIT_PERIOD)
                
                assertEqual(['/', '/assets/image.png', _FAVICON_PATH], server.requested_paths)
                server.requested_paths.clear()
                
                rr = revision_future.result()
                rr_body_filepath = rr._body_filepath  # capture
            
            # Simulate loss of revision body file, perhaps due to an
            # incomplete copy of a .crystalproj from one disk to another
            # (perhaps because of bad blocks in the revision body file)
            os.remove(rr_body_filepath)
            
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                with redirect_stderr(io.StringIO()) as captured_stderr:
                    r = Resource(project, server.get_url('/'))
                    revision_future = r.download(wait_for_embedded=True)
                    while not revision_future.done():
                        await bg_sleep(DEFAULT_WAIT_PERIOD)
                
                assert (
                    ' is missing its body on disk. Redownloading it.'
                    in captured_stderr.getvalue()
                )
                assertEqual(['/'], server.requested_paths)


async def test_when_download_resource_given_all_embedded_resources_already_downloaded_then_completes_early() -> None:
    server = MockHttpServer({
        '/': dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=dedent(
                """
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets/image.png" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8')
        ),
        '/index.php': dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=dedent(
                """
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets/image.png" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8')
        )
    })
    with server:
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, server.get_url('/'))
            revision_future = r.download(wait_for_embedded=True)
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            assertEqual(['/', '/assets/image.png', _FAVICON_PATH], server.requested_paths)
            server.requested_paths.clear()
            
            r = Resource(project, server.get_url('/index.php'))
            revision_future = r.download(wait_for_embedded=True)
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            assertEqual(['/index.php'], server.requested_paths)


async def test_given_same_resource_embedded_multiple_times_then_downloads_it_only_once() -> None:
    server = MockHttpServer({
        '/': dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=dedent(
                """
                <!DOCTYPE html>
                <html>
                <body>
                    <img src="/assets/image.png" />
                    <img src="/assets/image.png" />
                    <img src="/assets/image.png#fragment-should-be-ignored" />
                </body>
                </html>
                """
            ).lstrip('\n').encode('utf-8')
        )
    })
    with server:
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, server.get_url('/'))
            revision_future = r.download(wait_for_embedded=True)
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            assertEqual(['/', '/assets/image.png', _FAVICON_PATH], server.requested_paths)


# ------------------------------------------------------------------------------
# DownloadResourceGroupTask Tests

@skip('covered by: test_some_tasks_may_complete_immediately')
async def test_can_download_group_with_nothing_as_source() -> None:
    # See subtest: task_type='DownloadResourceGroupTask'
    pass


@skip('covered by: test_can_download_and_serve_a_static_site')
async def test_can_download_group_with_root_resource_as_source() -> None:
    # See section: "Test can download resource group, when root resource is source"
    pass


@skip('covered by: test_can_download_and_serve_a_static_site')
async def test_can_download_group_with_group_as_source() -> None:
    # See section: "Test can update membership of resource group, when other resource group is source"
    pass


async def test_can_download_empty_group() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        comic_pattern = sp.get_request_url('https://xkcd.com/#/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, _):
            assert mw.new_group_button.Enabled
            click_button(mw.new_group_button)
            ngd = await NewGroupDialog.wait_for()
            
            ngd.pattern_field.Value = comic_pattern
            ngd.name_field.Value = 'Comic'
            await ngd.ok()
            
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            comic_ti = root_ti.find_child(comic_pattern)
            comic_ti.SelectItem()
            
            class DownloadResourceGroupTaskSpy(crystal.task.DownloadResourceGroupTask):
                finish_call_count = 0
                
                def finish(self, *args, **kwargs):
                    DownloadResourceGroupTaskSpy.finish_call_count += 1
                    return super().finish(*args, **kwargs)
            
            with patch('crystal.task.DownloadResourceGroupTask', DownloadResourceGroupTaskSpy):
                click_button(mw.download_button)
                await wait_for(lambda: DownloadResourceGroupTaskSpy.finish_call_count >= 1 or None)
                assert 1 == DownloadResourceGroupTaskSpy.finish_call_count


# ------------------------------------------------------------------------------



from crystal.util.wx_bind import bind
from crystal.util.wx_date_picker import fix_date_picker_size
from crystal.util.wx_dialog import position_dialog_initially
from crystal.util.wx_static_box_sizer import wrap_static_box_sizer_child
from crystal.util.xos import is_linux, is_mac_os, is_windows
import datetime
from typing import Dict, TYPE_CHECKING
from tzlocal import get_localzone
import wx

if TYPE_CHECKING:
    from crystal.doc.html import HtmlParserType
    from crystal.model import Project


_WINDOW_INNER_PADDING = 10
_FORM_LABEL_INPUT_SPACING = 5
_FORM_ROW_SPACING = 10


class PreferencesDialog:
    _LXML_ITEM = \
        'Fastest - lxml'
    _HTML_PARSER_BS4_ITEM = \
        'Classic - html.parser (bs4)'
    
    _HTML_PARSER_TYPE_FOR_ITEM = {
        _LXML_ITEM: 'lxml',
        _HTML_PARSER_BS4_ITEM: 'html_parser',
    }  # type: Dict[str, HtmlParserType]
    _ITEM_FOR_HTML_PARSER_TYPE = {
        v: k for (k, v) in _HTML_PARSER_TYPE_FOR_ITEM.items()
    }
    
    # === Init ===
    
    def __init__(self, parent: wx.Window, project: 'Project') -> None:
        """
        Arguments:
        * parent -- parent wx.Window that this dialog is attached to.
        """
        self._project = project
        
        dialog = self.dialog = wx.Dialog(parent, title='Preferences', name='cr-preferences-dialog')
        dialog_sizer = wx.BoxSizer(wx.VERTICAL)
        dialog.SetSizer(dialog_sizer)
        bind(dialog, wx.EVT_BUTTON, self._on_button)
        bind(dialog, wx.EVT_CLOSE, self._on_close)
        
        project_box_sizer = wx.StaticBoxSizer(wx.VERTICAL, dialog, label='Project')
        project_box_sizer.Add(wrap_static_box_sizer_child(
            self._create_project_fields(project_box_sizer.GetStaticBox())))
        dialog_sizer.Add(project_box_sizer, flag=wx.EXPAND|wx.ALL,
            border=_WINDOW_INNER_PADDING)
        
        session_box_sizer = wx.StaticBoxSizer(wx.VERTICAL, dialog, label='Session')
        session_box_sizer.Add(wrap_static_box_sizer_child(
            self._create_session_fields(session_box_sizer.GetStaticBox())))
        dialog_sizer.Add(session_box_sizer, flag=wx.EXPAND|(wx.ALL & ~wx.TOP),
            border=_WINDOW_INNER_PADDING)
        
        dialog_sizer.Add(dialog.CreateButtonSizer(wx.OK|wx.CANCEL), flag=wx.EXPAND|wx.BOTTOM,
            border=_WINDOW_INNER_PADDING)
        
        position_dialog_initially(dialog)
        dialog.Show(True)
        dialog.Fit()  # NOTE: Must Fit() after Show() here so that wxGTK actually fits correctly
    
    def _create_project_fields(self, parent: wx.Window) -> wx.Sizer:
        fields_sizer = wx.GridBagSizer(
            vgap=_FORM_ROW_SPACING, hgap=_FORM_LABEL_INPUT_SPACING)
        
        fields_sizer.Add(
            wx.StaticText(parent, label='These preferences are saved with the project.'),
            flag=wx.EXPAND|wx.TOP, pos=wx.GBPosition(0, 0), span=wx.GBSpan(1, 2),
            border=5 if is_windows() else 0)
        
        fields_sizer.Add(
            wx.StaticText(parent, label='HTML Parser:'),
            flag=wx.EXPAND, pos=wx.GBPosition(1, 0))
        self.html_parser_field = wx.Choice(
            parent,
            name='cr-preferences-dialog__html-parser-field')
        self.html_parser_field.SetItems([
            self._LXML_ITEM,
            self._HTML_PARSER_BS4_ITEM,
        ])
        self._old_html_parser_type = self._project.html_parser_type
        self.html_parser_field.Selection = self.html_parser_field.Items.index(
            self._ITEM_FOR_HTML_PARSER_TYPE[self._old_html_parser_type])
        fields_sizer.Add(
            self.html_parser_field,
            flag=wx.EXPAND, pos=wx.GBPosition(1, 1))
        
        return fields_sizer
    
    def _create_session_fields(self, parent: wx.Window) -> wx.Sizer:
        fields_sizer = wx.GridBagSizer(
            vgap=_FORM_ROW_SPACING, hgap=_FORM_LABEL_INPUT_SPACING)
        
        fields_sizer.Add(
            wx.StaticText(parent, label='These preferences reset when Crystal quits.'),
            flag=wx.EXPAND|wx.TOP, pos=wx.GBPosition(0, 0), span=wx.GBSpan(1, 2),
            border=5 if is_windows() else 0)
        
        fields_sizer.Add(
            wx.StaticText(parent, label='To redownload URLs, provide:'),
            flag=wx.EXPAND|wx.TOP, pos=wx.GBPosition(1, 0), span=wx.GBSpan(1, 2),
            border=8)
        fields_sizer.Add(
            self._create_stale_before_field(parent),
            flag=wx.EXPAND, pos=wx.GBPosition(2, 0), span=wx.GBSpan(1, 2))
        
        fields_sizer.Add(
            wx.StaticText(parent, label='To download a site requiring login, provide:'),
            flag=wx.EXPAND, pos=wx.GBPosition(3, 0), span=wx.GBSpan(1, 2))
        fields_sizer.Add(
            wx.StaticText(parent, label='Cookie:'),
            flag=wx.EXPAND, pos=wx.GBPosition(4, 0))
        self.cookie_field = wx.ComboBox(
            parent,
            name='cr-preferences-dialog__cookie-field')
        self.cookie_field.SetItems(
            self._project.request_cookies_in_use(most_recent_first=True))
        self.cookie_field.Value = self._project.request_cookie or ''
        fields_sizer.Add(
            self.cookie_field,
            flag=wx.EXPAND, pos=wx.GBPosition(4, 1))
        
        return fields_sizer
    
    def _create_stale_before_field(self, parent: wx.Window) -> wx.Sizer:
        import wx.adv  # import late because does print spurious messages on macOS
        
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        
        self.stale_before_checkbox = wx.CheckBox(
            parent, label='Treat URLs as stale older than:',
            name='cr-preferences-dialog__stale-before-checkbox')
        bind(self.stale_before_checkbox, wx.EVT_CHECKBOX, self._update_stale_before_date_picker_enabled)
        sizer.Add(self.stale_before_checkbox, flag=wx.CENTER)
        
        self.stale_before_date_picker = wx.adv.DatePickerCtrl(
            parent, dt=wx.DefaultDateTime,  # default to today
            name='cr-preferences-dialog__stale-before-date-picker')
        sizer.Add(self.stale_before_date_picker, flag=wx.CENTER)
        fix_date_picker_size(self.stale_before_date_picker)
        
        stale_before_dt = self._project.min_fetch_date
        if stale_before_dt is not None:
            stale_before_dt_local = stale_before_dt.astimezone(tz=None)  # local timezone
            self.stale_before_checkbox.Value = True
            self.stale_before_date_picker.Value = wx.DateTime(
                year=stale_before_dt_local.year,
                month=stale_before_dt_local.month - 1,
                day=stale_before_dt_local.day)
        
        self._update_stale_before_date_picker_enabled()
        
        return sizer
    
    # === Events ===
    
    def _update_stale_before_date_picker_enabled(self, event: wx.CommandEvent | None=None) -> None:
        self.stale_before_date_picker.Enabled = self.stale_before_checkbox.Value
    
    def _on_button(self, event: wx.CommandEvent) -> None:
        btn_id = event.GetEventObject().GetId()
        if btn_id == wx.ID_OK:
            self._on_ok(event)
        elif btn_id == wx.ID_CANCEL:
            self._on_cancel(event)
    
    def _on_close(self, event: wx.CloseEvent) -> None:
        self._on_cancel(event)
    
    def _on_ok(self, event: wx.CommandEvent) -> None:
        # Save project fields
        new_html_parser_type = \
            self._HTML_PARSER_TYPE_FOR_ITEM[
                self.html_parser_field.Items[self.html_parser_field.Selection]]
        if new_html_parser_type != self._old_html_parser_type:
            self._project.html_parser_type = new_html_parser_type
        
        # Save session fields
        self._project.request_cookie = self.cookie_field.Value or None
        if self.stale_before_checkbox.Value:
            stale_before_wdt = self.stale_before_date_picker.Value
            stale_before_dt = datetime.datetime(
                year=stale_before_wdt.year,
                month=stale_before_wdt.month + 1,
                day=stale_before_wdt.day)
            stale_before_dt_local = stale_before_dt.replace(tzinfo=get_localzone())
            self._project.min_fetch_date = stale_before_dt_local
        else:
            self._project.min_fetch_date = None
        
        self.dialog.Destroy()
    
    def _on_cancel(self, event: wx.Event) -> None:
        self.dialog.Destroy()



from collections.abc import Callable, Iterator
from contextlib import contextmanager, redirect_stdout
import io
import sys
from typing import cast, TextIO


@contextmanager
def console_output_copied() -> Iterator[io.StringIO]:
    buffer = io.StringIO()
    stdout_and_buffer = cast(TextIO, _MultiplexedWriteOnlyTextIO([sys.stdout, buffer]))
    with redirect_stdout(stdout_and_buffer):
        yield buffer


_WRITE_CALLABLE_ATTRS = [
    # IOBase
    'close', 'flush', 'seek', 'truncate', 'writelines',
    # BufferedIOBase
    'write',
    # TextIOBase
    # (no additional attributes)
]

class _MultiplexedWriteOnlyTextIO:
    def __init__(self, bases: list[TextIO]) -> None:
        if not len(bases) >= 1:
            raise ValueError()
        self._bases = bases
    
    def __getattr__(self, attr_name: str):
        if attr_name in _WRITE_CALLABLE_ATTRS:
            return _MultiplexedCallable([
                getattr(base, attr_name)
                for base in self._bases
            ])
        else:
            return getattr(self._bases[0], attr_name)


class _MultiplexedCallable:
    def __init__(self, bases: list[Callable]) -> None:
        if not len(bases) >= 1:
            raise ValueError()
        self._bases = bases
    
    def __call__(self, *args, **kwargs):
        result = self._bases[0](*args, **kwargs)
        for other_base in self._bases[1:]:
            other_base(*args, **kwargs)
        return result



from collections.abc import Iterator
from contextlib import contextmanager
import crystal.task
from crystal.tests.util.tasks import scheduler_thread_context
import socket
from typing import NoReturn, TYPE_CHECKING
from unittest.mock import patch

if TYPE_CHECKING:
    from crystal.task import DownloadResourceGroupTask


@contextmanager
def delay_between_downloads_minimized() -> Iterator[None]:
    old_value = crystal.task.DELAY_BETWEEN_DOWNLOADS
    # NOTE: Must be long enough so that download tasks stay around long enough
    #       to be observed, but short enough to provide a speed boost
    crystal.task.DELAY_BETWEEN_DOWNLOADS = 0.2
    try:
        yield
    finally:
        crystal.task.DELAY_BETWEEN_DOWNLOADS = old_value


def load_children_of_drg_task(
        drg_task: 'DownloadResourceGroupTask',
        *, scheduler_thread_enabled: bool=True,
        task_added_to_project: bool=True,
        ) -> None:
    if not (scheduler_thread_enabled == False or
            task_added_to_project == False):
        raise ValueError(
            'The specified task must not be scheduled on a '
            'running scheduler thread')
    
    # Precondition
    assert not drg_task._download_members_task._children_loaded
    
    with scheduler_thread_context():
        task_unit = drg_task._download_members_task.try_get_next_task_unit()
        assert task_unit is not None
        task_unit()
    
    # Postcondition
    assert drg_task._download_members_task._children_loaded


@contextmanager
def network_down() -> Iterator[None]:
    def MockHTTPConnection(*args, **kwargs) -> NoReturn:
        raise socket.gaierror(8, 'nodename nor servname provided, or not known')
    
    with patch('crystal.download.HTTPConnection', MockHTTPConnection):
        yield



from __future__ import annotations

from collections.abc import Callable
import sqlite3
from typing import Self

# Whether to print each database query
VERBOSE_QUERIES = False


class DatabaseConnection:
    """Wraps a sqlite3.dbapi2.Connection, ensuring that it is used correctly."""
    
    def __init__(self,
            db: sqlite3.dbapi2.Connection,
            readonly_func: Callable[[], bool]) -> None:
        self._db = db
        self._readonly_func = readonly_func
    
    def cursor(self, *args, **kwargs) -> DatabaseCursor:
        c = self._db.cursor(*args, **kwargs)  # type: sqlite3.dbapi2.Cursor
        return DatabaseCursor(c, self._readonly_func())
    
    def __getattr__(self, attr_name: str):
        return getattr(self._db, attr_name)


class DatabaseCursor:
    """Wraps a sqlite3.dbapi2.Cursor, ensuring that it is used correctly."""
    
    def __init__(self, c: sqlite3.dbapi2.Cursor, readonly: bool) -> None:
        self._c = c
        self._readonly = readonly
    
    def execute(self, command: str, *args, **kwargs) -> Self:
        ignore_readonly = bool(kwargs.pop('ignore_readonly', False))
        
        # Ensure that caller does disallow commands that write to the database
        # if the project is readonly
        if self._readonly and not ignore_readonly:
            command_lower = command.lower()  # cache
            command_is_read = (
                command_lower.startswith('select ') or
                command_lower.startswith('pragma table_info(') or
                command_lower.startswith('explain ')
            )
            command_is_write = not command_is_read  # conservative
            if command_is_write:
                raise AssertionError(
                    'Attempted to write to database when (Project.readonly == True). '
                    'Caller should have checked this case and thrown ProjectReadOnlyError.'
                )
        
        if VERBOSE_QUERIES:
            print(f'QUERY: {command!r}, {args=}')
        result = self._c.execute(command, *args, **kwargs)
        assert result is self._c
        return self
    
    # Define specially to help mypy know that this attribute exists
    def __iter__(self, *args, **kwargs):
        return self._c.__iter__(*args, **kwargs)
    
    def __getattr__(self, attr_name: str):
        return getattr(self._c, attr_name)


def get_column_names_of_table(c: DatabaseCursor, table_name: str) -> list[str]:
    return [
        column_name
        for (_, column_name, column_type, _, _, _)
        # NOTE: Cannot use regular '?' placeholder in this PRAGMA
        in c.execute('PRAGMA table_info({})'.format(table_name))
    ]


def is_no_such_column_error_for(column_name: str, e: Exception) -> bool:
    return (
        isinstance(e, sqlite3.OperationalError) and
        str(e) == f'no such column: {column_name}'
    )


def get_index_names(c: DatabaseCursor) -> list[str]:
    return [
        index_name
        for (index_name,) in 
        c.execute('SELECT name FROM sqlite_master WHERE type = "index"')
    ]



from typing import TypeVar

_E = TypeVar('_E')


def dedup_list(xs: list[_E]) -> list[_E]:
    """
    Removes duplicates from the specified list,
    preserving the original order of elements.
    """
    return list(dict.fromkeys(xs))



from crystal.model import Project, ResourceGroup
from crystal.tests.util.controls import click_button, TreeItem
from crystal.tests.util.server import extracted_project
from crystal.tests.util.wait import (
    first_child_of_tree_item_is_not_loading_condition, is_enabled_condition,
    wait_for,
)
from crystal.tests.util.windows import NewGroupDialog, OpenOrCreateDialog
from crystal.util.wx_dialog import mocked_show_modal
from crystal.util.xos import is_windows
from unittest import skip
from unittest.mock import patch
import wx


async def test_can_edit_name_of_group() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        comic_pattern = 'https://xkcd.com/#/'
        
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            
            home_ti = root_ti.GetFirstChild()
            assert home_ti is not None
            assert home_ti.Text.endswith('Home')
            
            home_ti.Expand()
            await wait_for(first_child_of_tree_item_is_not_loading_condition(home_ti))
            home_ti__comic_ti = home_ti.find_child(comic_pattern)
            assert home_ti__comic_ti.Text.endswith('Comics')
            
            comic_ti = root_ti.find_child(comic_pattern)
            assert comic_ti.Text.endswith('Comics')
            
            # Ensure can rename top-level ResourceGroupNode
            if True:
                comic_ti.SelectItem()
                
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                ngd = await NewGroupDialog.wait_for()
                
                # 1. Ensure prepopulates correct information
                # 2. Test: Cannot edit URL pattern
                assert comic_pattern == ngd.pattern_field.Value
                assert 'Comics' == ngd.name_field.Value
                assert 'Home' == ngd.source
                assert not ngd.pattern_field.Enabled
                assert ngd.name_field.HasFocus  # default focused field
                
                ngd.name_field.Value = 'Comics2'
                await ngd.ok()
                comic_ti = root_ti.find_child(comic_pattern)
                assert comic_ti.Text.endswith('Comics2')
                
                # Ensure selection did not change
                assert comic_ti.IsSelected()
                
                # Ensure all copies of comic node have updated name
                assert home_ti.IsExpanded()
                home_ti__comic_ti = home_ti.find_child(comic_pattern)
                assert home_ti__comic_ti.Text.endswith('Comics2')
            
            # Ensure can rename nested GroupedLinkedResourcesNode
            if True:
                home_ti__comic_ti.SelectItem()
                
                if is_windows():
                    await wait_for(is_enabled_condition(mw.edit_button), timeout=.5)
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                ngd = await NewGroupDialog.wait_for()
                
                # Ensure prepopulates correct information
                assert comic_pattern == ngd.pattern_field.Value
                assert 'Comics2' == ngd.name_field.Value
                
                ngd.name_field.Value = 'Comics'
                await ngd.ok()
                home_ti = root_ti.GetFirstChild()
                assert home_ti is not None
                
                # Ensure selection did not change
                assert home_ti.IsExpanded()
                home_ti__comic_ti = home_ti.find_child(comic_pattern)
                assert home_ti__comic_ti.IsSelected()
                
                # Ensure all copies of home node have updated name
                comic_ti = root_ti.find_child(comic_pattern)
                assert comic_ti.Text.endswith('Comics')
                assert home_ti__comic_ti.Text.endswith('Comics')


async def test_can_edit_source_of_group() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        comic_pattern = 'https://xkcd.com/#/'
        first_comic_pattern = 'https://xkcd.com/1/'
        
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            comic_ti = root_ti.find_child(comic_pattern)
            comic_ti.SelectItem()
            
            # Ensure can edit source to be root resource
            if True:
                assert comic_ti.IsSelected()
                
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                ngd = await NewGroupDialog.wait_for()
                assert 'Home' == ngd.source
                
                ngd.source = 'Atom Feed'
                await ngd.ok()
                
                click_button(mw.edit_button)
                ngd = await NewGroupDialog.wait_for()
                assert 'Atom Feed' == ngd.source
                await ngd.ok()
            
            # Ensure can edit source to be group
            if True:
                ResourceGroup(project, 'First Comic', first_comic_pattern)
                first_comic_ti = root_ti.find_child(first_comic_pattern)
                
                assert comic_ti.IsSelected()
                
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                ngd = await NewGroupDialog.wait_for()
                assert 'Atom Feed' == ngd.source
                
                ngd.source = 'First Comic'
                await ngd.ok()
                
                click_button(mw.edit_button)
                ngd = await NewGroupDialog.wait_for()
                assert 'First Comic' == ngd.source
                await ngd.ok()
            
            # Ensure cannot edit source to be self
            if True:
                assert comic_ti.IsSelected()
                
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                ngd = await NewGroupDialog.wait_for()
                assert 'First Comic' == ngd.source
                
                ngd.source = 'Comics'
                
                with patch(
                        'crystal.browser.new_group.ShowModal',
                        mocked_show_modal('cr-source-cycle-created', wx.ID_OK)
                        ) as show_modal_method:
                    click_button(ngd.ok_button)
                    assert 1 == show_modal_method.call_count
                
                await ngd.cancel()
            
            # Ensure cannot edit source to be group that would create cycle
            if True:
                first_comic_ti.SelectItem()
                
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                ngd = await NewGroupDialog.wait_for()
                assert None == ngd.source
                
                ngd.source = 'Comics'
                
                with patch(
                        'crystal.browser.new_group.ShowModal',
                        mocked_show_modal('cr-source-cycle-created', wx.ID_OK)
                        ) as show_modal_method:
                    click_button(ngd.ok_button)
                    assert 1 == show_modal_method.call_count
                
                await ngd.cancel()


@skip('covered by: test_can_edit_name_of_group')
async def test_cannot_edit_url_pattern_of_group() -> None:
    pass


@skip('covered by: test_can_edit_source_of_group')
async def test_given_editing_group_when_select_self_as_source_and_press_save_then_shows_error_dialog() -> None:
    pass


@skip('covered by: test_can_edit_source_of_group')
async def test_given_editing_group_when_select_source_that_would_create_cycle_and_press_save_then_shows_error_dialog() -> None:
    pass



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from copy import deepcopy
from crystal import resources
from crystal.model import Project
from crystal.server import get_request_url, ProjectServer
from crystal.tests.util.runner import bg_fetch_url
from crystal.tests.util.wait import DEFAULT_WAIT_TIMEOUT
from crystal.util import http_date
from crystal.util.bulkheads import capture_crashes_to_stderr
from crystal.util.xdatetime import datetime_is_aware
from crystal.util.xthreading import bg_call_later, fg_call_and_wait
import datetime
from email.message import EmailMessage
from http.server import BaseHTTPRequestHandler, HTTPServer
import os
import re
import tempfile
from typing import List, Optional
import unittest.mock
from zipfile import ZipFile

# ------------------------------------------------------------------------------
# Utility: Server

@contextmanager
def served_project(
        zipped_project_filename: str,
        **kwargs,
        ) -> Iterator[ProjectServer]:
    with extracted_project(zipped_project_filename) as project_dirpath:
        with served_project_from_filepath(project_dirpath, **kwargs) as sp:
            yield sp


@contextmanager
def served_project_from_filepath(
        project_dirpath: str,
        *, fetch_date_of_resources_set_to: datetime.datetime | None=None,
        port: int | None=None,
        ) -> Iterator[ProjectServer]:
    if fetch_date_of_resources_set_to is not None:
        if not datetime_is_aware(fetch_date_of_resources_set_to):
            raise ValueError('Expected fetch_date_of_resources_set_to to be an aware datetime')
    
    must_alter_fetch_date = (fetch_date_of_resources_set_to is not None)
    project_server = None  # type: Optional[ProjectServer]
    project = fg_call_and_wait(lambda: Project(project_dirpath, readonly=True if not must_alter_fetch_date else False))
    try:
        # Alter the fetch date of every ResourceRevision in the project
        # to match "fetch_date_of_resources_set_to", if provided
        if must_alter_fetch_date:
            def fg_task() -> None:
                for r in project.resources:
                    for rr in list(r.revisions()):
                        if rr.metadata is None:
                            print(
                                f'Warning: Unable to alter fetch date of '
                                f'resource revision lacking HTTP headers: {rr}')
                            continue
                        
                        assert fetch_date_of_resources_set_to is not None
                        rr_new_date = http_date.format(fetch_date_of_resources_set_to)
                        
                        # New Metadata = Old Metadata with Date and Age headers replaced
                        rr_new_metadata = deepcopy(rr.metadata)
                        rr_new_metadata['headers'] = [
                            [cur_name, cur_value]
                            for (cur_name, cur_value) in
                            rr_new_metadata['headers']
                            if cur_name.lower() not in ['date', 'age']
                        ] + [['Date', rr_new_date]]
                        
                        rr._alter_metadata(rr_new_metadata, ignore_readonly=True)
            fg_call_and_wait(fg_task)
        
        # Start server
        project_server = ProjectServer(project,
            port=(port or 2798),  # CRYT on telephone keypad
            verbosity='indent',
        )
        yield project_server
    finally:
        def close_project() -> None:
            if project_server is not None:
                project_server.close()
            project.close()
        fg_call_and_wait(close_project)


class MockHttpServer:
    def __init__(self, routes) -> None:
        self.requested_paths = []  # type: List[str]
        
        mock_server = self  # capture
        class RequestHandler(BaseHTTPRequestHandler):
            def do_GET(self) -> None:  # override
                mock_server.requested_paths.append(self.path)
                
                # Look for static route
                route = routes.get(self.path)
                if route is not None:
                    self._send_route_response(route)
                    return
                
                # Look for dynamic route
                for (path_pattern, route) in routes.items():
                    if not callable(path_pattern):
                        continue
                    if not path_pattern(self.path):
                        continue
                    self._send_route_response(route)
                    return
                
                self.send_response(404)
                self.end_headers()
                return
            
            def _send_route_response(self, route) -> None:
                self.send_response(route['status_code'])
                for (k, v) in route['headers']:
                    self.send_header(k, v)
                self.end_headers()
                
                content = route['content']
                if callable(content):
                    content = content(self.path)
                assert isinstance(content, bytes)
                self.wfile.write(content)
        
        self._port = 2798  # CRYT on telephone keypad
        address = ('', self._port)
        self._server = HTTPServer(address, RequestHandler)
        
        @capture_crashes_to_stderr
        def bg_task() -> None:
            try:
                self._server.serve_forever()
            finally:
                self._server.server_close()
        bg_call_later(bg_task, daemon=True)
    
    def get_url(self, path: str) -> str:
        return f'http://localhost:{self._port}' + path
    
    def close(self) -> None:
        self._server.shutdown()
    
    def __enter__(self) -> MockHttpServer:
        return self
    
    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.close()


# ------------------------------------------------------------------------------
# Utility: Extracted Project

@contextmanager
def extracted_project(
        zipped_project_filename: str
        ) -> Iterator[str]:
    # NOTE: If a file inside the temporary directory is still open,
    #       ignore_cleanup_errors=True will prevent Windows from raising,
    #       at the cost of leaving the temporary directory around
    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as project_parent_dirpath:
        # Extract project
        with resources.open_binary(zipped_project_filename) as zipped_project_file:
            with ZipFile(zipped_project_file, 'r') as project_zipfile:
                project_zipfile.extractall(project_parent_dirpath)
        
        # Open project
        (project_filename,) = (
            fn for fn in os.listdir(project_parent_dirpath)
            if fn.endswith('.crystalproj')
        )
        project_dirpath = os.path.join(project_parent_dirpath, project_filename)
        yield project_dirpath


# ------------------------------------------------------------------------------
# Utility: Server Requests

@contextmanager
def assert_does_open_webbrowser_to(request_url: str) -> Iterator[None]:
    with unittest.mock.patch('webbrowser.open', spec=True) as mock_open:
        yield
        mock_open.assert_called_with(request_url)


async def is_url_not_in_archive(archive_url: str) -> bool:
    server_page = await fetch_archive_url(
        archive_url, 
        headers={'X-Crystal-Dynamic': 'False'})
    return server_page.is_not_in_archive


async def fetch_archive_url(
        archive_url: str,
        port: int | None=None,
        *, headers: dict[str, str] | None=None,
        timeout: float | None=None,
        ) -> WebPage:
    if timeout is None:
        timeout = DEFAULT_WAIT_TIMEOUT
    return await bg_fetch_url(get_request_url(archive_url, port), headers=headers, timeout=timeout)


# TODO: Rename as HttpResponse or ProjectServerResponse
class WebPage:
    def __init__(self, status: int, headers: EmailMessage, content_bytes: bytes) -> None:
        self._status = status
        self._headers = headers
        self._content_bytes = content_bytes
        self._content = None  # type: Optional[str]
    
    # === High-Level Attributes ===
    
    @property
    def is_not_in_archive(self) -> bool:
        return (
            self._status == 404 and
            self.title == 'Not in Archive | Crystal'
        )
    
    @property
    def etag(self) -> str | None:
        return self._headers.get('ETag')
    
    @property
    def title(self) -> str | None:
        # TODO: Use an HTML parser to improve robustness
        m = re.search(r'<title>([^<]*)</title>', self.content)
        if m is None:
            return None
        else:
            return m.group(1).strip()
    
    # === Low-Level Attributes ===
    
    @property
    def status(self) -> int:
        return self._status
    
    @property
    def headers(self) -> EmailMessage:
        return self._headers
    
    @property
    def content(self) -> str:  # lazy
        if self._content is None:
            self._content = self._content_bytes.decode('utf-8')
        return self._content
    
    @property
    def content_bytes(self) -> bytes:
        return self._content_bytes


# ------------------------------------------------------------------------------
